/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: okbc_get.c,v 1.10 2004/10/18 14:22:14 theno23 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rdfsql/rdfsql.h"
#include "okbc.h"

/* okbc_get_behavior_supported_values(behavior, kb)
 *
 * (behavior &key kb)
 *
 * Return Value(s): behavior-values 
 * 
 * Flags: M R 
 * 
 * Returns a list of the supported values of the behavior 
 * the KB is capable of supporting. For example, the KB might support both
 * the :immediate and :never variants of the behavior
 * :constraint-checking-time. These two options would be returned
 * as a list. The returned value behavior-values is always a list,
 * even when no variants are supported - that is, it is ().
 */

okbc_node *okbc_get_behavior_supported_values(const char *behavior,
					      const char *kb)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_behavior_values(behavior, kb)
 *
 * (behavior &key kb)
 *
 * Return Value(s): behavior-values 
 * 
 * Flags: M R 
 * 
 * Returns a list of active values of the behavior under
 * which the KB is currently operating. For example, the KB might support
 * both the :immediate and :never variants of the behavior
 * :constraint-checking-time, but only one of these modes can be
 * enabled at a given time. A list containing, for example, just
 * :never would be returned.
 * The returned value behavior-values is always a list, even when
 * no variants are active - that is, it is the ().
 */

okbc_node *okbc_get_behavior_values(const char *behavior, const char *kb)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_class_instances(class, kb, inference_level, number_of_values,
 * kb_local_only_p)
 *
 * (class &key kb (inference-level :taxonomic) (number-of-values :all)
 * kb-local-only-p)
 *
 * Return Value(s): list-of-instances exact-p more-status 
 * 
 * Flags: E M R 
 * 
 * Returns a list-of-instances for class.
 */

okbc_node *okbc_get_class_instances(const char *class, const char *kb,
				    const char *inference_level,
				    int number_of_values,
				    int kb_local_only_p)
{
    okbc_node *ret = NULL;
    okbc_node *inst = NULL;
    rs_result *classes, *instances;
    rs_triple *t, *u;
    GSList *list = NULL, *it;
    int i = 0;

    ret = okbc_new_list_node(3);

    if (!inference_level || inference_level == key_taxonomic) {
        /* Find the instances of this class and all its subclasses */
        classes = rs_find_triples(okbc_intl_db, NULL, RDFS_SUBCLASS, class, ObjURI, 0, kb);

        while ((t = rs_next_triple(classes))) {
	    instances = rs_find_triples(okbc_intl_db, NULL, RDF_TYPE,
			    t->subject, ObjURI, number_of_values, kb);
	    if (number_of_values) {
	        while ((u = rs_next_triple(instances))) {
		    if (i++ < number_of_values) {
		        list = g_slist_append(list, strdup(u->subject));
		    }
	        }
	        if (i >= number_of_values) {
		    /* Empty the results object */
		    while ((t = rs_next_triple(classes))) { ; }
		    break;
	        }
	    } else {
	        while ((u = rs_next_triple(instances))) {
		    list = g_slist_append(list, strdup(u->subject));
                }
	    }
	    rs_free_result(instances);
	}
	rs_free_result(classes);
	inst = okbc_new_list_node(g_slist_length(list));
	i=0;
	for (it = list; it; it = it->next) {
		inst->data.list->items[i++] = okbc_new_node(OKBC_URI,
				okbc_new_atom(it->data));
		free(it->data);
	}
	g_slist_free(list);
    } else {
	/* We just want direct instances of this class */
        instances = rs_find_asserted_triples(okbc_intl_db, NULL, RDF_TYPE, class, ObjURI, number_of_values, kb);

	inst = okbc_new_list_node(rs_num_triples(instances));
	while ((u = rs_next_triple(instances))) {
		inst->data.list->items[i++] = okbc_new_node(OKBC_URI,
				okbc_new_atom(u->subject));
	}
	rs_free_result(instances);
    }
	
    ret->data.list->items[0] = inst;
    ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));
    ret->data.list->items[2] = okbc_new_node(OKBC_BOOL, ATOM(0));

    return ret;
}

/* okbc_get_class_subclasses(class, kb, inference_level, number_of_values,
 * kb_local_only_p)
 *
 * (class &key kb (inference-level :taxonomic) (number-of-values :all)
 * kb-local-only-p)
 *
 * Return Value(s): list-of-subclasses exact-p more-status 
 * 
 * Flags: E M R 
 * 
 * Returns the list-of-subclasses of class.
 */

okbc_node *okbc_get_class_subclasses(const char *class, const char *kb,
				     const char *inference_level,
				     int number_of_values,
				     int kb_local_only_p)
{
    okbc_node *ret = NULL;
    okbc_node *inst = NULL;
    rs_result *classes;
    rs_triple *t;
    int i;

    ret = okbc_new_list_node(3);

    if (!inference_level || inference_level == key_taxonomic) {
        classes = rs_find_triples(okbc_intl_db, NULL,
                                  RDFS_SUBCLASS, class,
                                  ObjURI, number_of_values, kb);

	inst = okbc_new_list_node(rs_num_triples(classes));
	i=0;
        while ((t = rs_next_triple(classes))) {
	    inst->data.list->items[i++] = okbc_new_node(OKBC_URI,
			okbc_new_atom(t->subject));
	}
	rs_free_result(classes);
    } else {
	/* We just want direct subclasses of this class */
        classes = rs_find_asserted_triples(okbc_intl_db, NULL,
                                           RDFS_SUBCLASS, class,
			ObjURI, number_of_values, kb);

	i=0;
	inst = okbc_new_list_node(rs_num_triples(classes));
	while ((t = rs_next_triple(classes))) {
		inst->data.list->items[i++] = okbc_new_node(OKBC_URI,
				okbc_new_atom(t->subject));
	}
	rs_free_result(classes);
    }
	
    ret->data.list->items[0] = inst;
    ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));
    ret->data.list->items[2] = okbc_new_node(OKBC_BOOL, ATOM(0));

    return ret;
}

/* okbc_get_class_superclasses(class, kb, inference_level, number_of_values,
 * kb_local_only_p)
 *
 * (class &key kb (inference-level :taxonomic) (number-of-values :all)
 * kb-local-only-p)
 *
 * Return Value(s): list-of-superclasses exact-p more-status 
 * 
 * Flags: E M R 
 * 
 * Returns the list-of-superclasses of class.
 */

okbc_node *okbc_get_class_superclasses(const char *class, const char *kb,
				       const char *inference_level,
				       int number_of_values,
				       int kb_local_only_p)
{
    okbc_node *ret = NULL;
    okbc_node *inst = NULL;
    rs_result *classes;
    rs_triple *t;
    int i;

    ret = okbc_new_list_node(3);

    if (!inference_level || inference_level == key_taxonomic) {
        classes = rs_find_triples(okbc_intl_db, class, RDFS_SUBCLASS, NULL, ObjURI, number_of_values, kb);

	i=0;
        inst = okbc_new_list_node(rs_num_triples(classes));
        while ((t = rs_next_triple(classes))) {
	    inst->data.list->items[i++] = okbc_new_node(OKBC_URI,
                        okbc_new_atom(t->object));
	}
	rs_free_result(classes);
    } else {
	/* We just want direct subclasses of this class */
        classes = rs_find_asserted_triples(okbc_intl_db, class, RDFS_SUBCLASS, NULL, ObjURI, number_of_values, kb);

	i=0;
	inst = okbc_new_list_node(rs_num_triples(classes));
	while ((t = rs_next_triple(classes))) {
		inst->data.list->items[i++] = okbc_new_node(OKBC_URI,
				okbc_new_atom(t->object));
	}
	rs_free_result(classes);
    }
	
    ret->data.list->items[0] = inst;
    ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));
    ret->data.list->items[2] = okbc_new_node(OKBC_BOOL, ATOM(0));

    return ret;
}

/* okbc_get_classes_in_domain_of(slot, kb, frame_attachment, inference_level,
 * kb_local_only_p)
 *
 * (slot &key kb frame-attachment (inference-level :taxonomic) kb-local-only-p)
 *
 * Return Value(s): classes 
 * 
 * Flags: O R 
 * 
 * Returns a list of classes that are known to be in the domain of slot with
 * respect to frame-attachment (if supplied). If frame-attachment is supplied,
 * it may be used as a hint to the KRS to limit the amount of computation
 * performed by constraining the search for classes to the superclasses or
 * types of frame-attachment.  Each class returned (and any subclass) is
 * guaranteed to be a legal frame argument for a slot operation on slot with
 * slot-type :template, e.g., put-slot-values.
 */

okbc_node *okbc_get_classes_in_domain_of(const char *slot, const char *kb,
					 const char *frame_attachment,
					 const char *inference_level,
					 int kb_local_only_p)
{
    okbc_node *ret = NULL;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    if (!inference_level || inference_level == key_taxonomic) {
	r = rs_find_triples(okbc_intl_db, slot, RDFS_DOMAIN, NULL, ObjURI,
                                 0, NULL);
    } else {
	r = rs_find_asserted_triples(okbc_intl_db, slot, RDFS_DOMAIN, NULL,
                                     ObjURI, 0, NULL);
    }

    ret = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
        ret->data.list->items[i++] = okbc_new_node(OKBC_URI,
		    okbc_new_atom(t->object));
    }
    rs_free_result(r);

    return ret;
}

/* okbc_get_facet_value(frame, slot, facet, kb, inference_level, slot_type, value_selector, kb_local_only_p)
 *
 * (frame slot facet &key kb (inference-level :taxonomic) (slot-type :own) (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): value-or-false exact-p 
 * 
 * Flags: O R 
 * 
 * Returns the sole member of the set of values
 * of the specified facet. It is most commonly used when that set is 
 * expected to have only one member. When the facet has no value,
 * value-or-false is false. It is an error to
 * call this operation on a non-single-valued facet; if it is called, a
 * cardinality-violation error should be signaled.
 */

okbc_node *okbc_get_facet_value(const char *frame, const char *slot,
				const char *facet, const char *kb,
				const char *inference_level,
				const char *slot_type,
				const char *value_selector,
				int kb_local_only_p)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_facet_values(frame, slot, facet, kb, inference_level, slot_type, number_of_values, value_selector, kb_local_only_p)
 *
 * (frame slot facet &key kb (inference-level :taxonomic) (slot-type :own) (number-of-values :all) (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): values exact-p more-status 
 * 
 * Flags: E O R 
 * 
 * Returns the set of values of the specified facet,
 * in no guaranteed order. It always returns a (possibly empty) list
 * of values.
 */

okbc_node *okbc_get_facet_values(const char *frame, const char *slot,
				 const char *facet, const char *kb,
				 const char *inference_level,
				 const char *slot_type,
				 int number_of_values,
				 const char *value_selector,
				 int kb_local_only_p)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_facet_values_in_detail(frame, slot, facet, kb, inference_level, slot_type, number_of_values, value_selector, kb_local_only_p)
 *
 * (frame slot facet &key kb (inference-level :taxonomic) (slot-type :own) (number-of-values :all) (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): list-of-specs exact-p more-status default-p 
 * 
 * Flags: E M R 
 * 
 * Returns the list-of-specs describing the values
 * of the facet of slot within frame, in no guaranteed
 * order. It always returns a list of specifications as values. If the
 * specified slot has no values, () is returned.
 *
 * Each spec is a 3-tuple of the form (value direct-p default-p).
 * 
 * value - a value of the facet
 * direct-p - a flag that is true if the value is known to be
 * directly asserted for the facet in the frame and
 * false otherwise
 * default-p - a flag that is true if the value is known to be
 * a default value, and false otherwise
 * The fourth returned value (default-p) is true if the
 * list-of-specs is (), and the fact that there are no values
 * is itself a default.
 */

okbc_node *okbc_get_facet_values_in_detail(const char *frame,
					   const char *slot,
					   const char *facet,
					   const char *kb,
					   const char *inference_level,
					   const char *slot_type,
					   int number_of_values,
					   const char *value_selector,
					   int kb_local_only_p)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_frame_details(frame, kb, inference_level, number_of_values,
 * kb_local_only_p)
 *
 * (frame &key kb (inference-level :taxonomic) (number-of-values :all)
 * kb-local-only-p)
 *
 * Return Value(s): details exact-p 
 * 
 * Flags: O R 
 * 
 * Returns a property list (list of alternating keywords and values)
 * describing the frame. The properties of the frame are computed
 * using the inference-level, number-of-values-p, and
 * kb-local-only-p arguments, whenever applicable to the appropriate
 * OKBC operator used to compute any given property. The set of properties
 * computed is at least the following:
 *
 * Property
 * Operation(s) used to compute property
 *
 * :handle
 * get-frame-handle
 *
 * :name
 * get-frame-name
 *
 * :pretty-name
 * get-frame-pretty-name
 *
 * :frame-type
 * get-frame-type
 *
 * :primitive-p
 * primitive-p
 *
 * :superclasses
 * get-class-superclasses
 *
 * :subclasses
 * get-class-subclasses
 *
 * :types
 * get-instance-types
 *
 * :own-slots
 * get-frame-slots, get-slot-values
 *
 * :template-slots
 * get-frame-slots, get-slot-values
 *
 * :own-facets
 * get-frame-slots, get-slot-values, get-slot-facets, get-facet-values
 *
 * :template-facets
 * get-frame-slots, get-slot-values, get-slot-facets, get-facet-values
 *
 * :sentences
 * get-frame-sentences
 *
 * The :own-slots, :own-facets, :template-slots, and
 * :template-facets properties returned are slot and facet
 * specifications as defined for create-frame. This operation is most
 * useful in low-bandwidth or high-latency applications. A single call to
 * get-frame-details is often sufficient to display all the
 * interesting features of a frame.
 */

okbc_node *okbc_get_frame_details(const char *frame, const char *kb,
				  const char *inference_level,
				  int number_of_values,
				  int kb_local_only_p)
{
    okbc_node *ret = okbc_new_list_node(19);
    okbc_node **items = ret->data.list->items;
    int i = 0;

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":get-frame-handle"));
    items[i++] = okbc_get_frame_handle(frame, kb, kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":get-frame-name"));
    items[i++] = okbc_get_frame_name(frame, kb, kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":get-frame-pretty-name"));
    items[i++] = okbc_get_frame_pretty_name(frame, kb, kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":get-frame-type"));
    items[i++] = okbc_get_frame_type(frame, kb, kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":get-class-superclasses"));
    items[i++] = okbc_get_class_superclasses(frame, kb, inference_level,
		    number_of_values, kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":get-class-subclasses"));
    items[i++] = okbc_get_class_subclasses(frame, kb, inference_level,
		    number_of_values, kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":get-instance-types"));
    items[i++] = okbc_get_instance_types(frame, kb, inference_level,
		    number_of_values, kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":own-slots"));
    items[i++] = okbc_get_frame_slots(frame, kb, inference_level,
		    NULL, kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_SYMBOL, ATOM(":sentences"));
    items[i++] = okbc_get_frame_sentences(frame, kb, number_of_values, 0, NULL,
		    kb_local_only_p);

    items[i++] = okbc_new_node(OKBC_BOOL, ATOM(1));

    return ret;
}

/* okbc_get_frame_handle(frame, kb, kb_local_only_p)
 *
 * (frame &key kb kb-local-only-p)
 *
 * Return Value(s): frame-handle 
 * 
 * Flags: O R 
 * 
 * Returns a frame-handle that uniquely identifies the frame
 * argument in kb.
 */

okbc_node *okbc_get_frame_handle(const char *frame, const char *kb,
				 int kb_local_only_p)
{
    char tmp[18];

    sprintf(tmp, "%llX", rs_hash(frame));

    return okbc_new_node(OKBC_LITERAL, okbc_new_atom(tmp));
}

/* okbc_get_frame_in_kb(thing, kb, error_p, kb_local_only_p)
 *
 * (thing &key kb (error-p true) kb-local-only-p)
 *
 * Return Value(s): frame frame-found-p 
 * 
 * Flags: M R 
 * 
 * Returns two values. The first value is the frame
 * identified by thing if such a frame is found, or false.
 * The second value (frame-found-p) is true iff thing is
 * coercible to a frame, and that frame is resident in KB. In
 * all cases it is verified that the frame does, in fact, reside in kb.
 * Otherwise, the frame-found-p value is nil (unless 
 * error-p is true, in which case the operation signals a
 * not-coercible-to-frame error because
 * thing is not a valid frame in kb).
 */

okbc_node *okbc_get_frame_in_kb(const char *thing, const char *kb,
				int error_p, int kb_local_only_p)
{
    okbc_node *ret = okbc_new_list_node(2);
    rs_result *r;

    r = rs_find_triples(okbc_intl_db, thing, NULL, NULL, ObjURI, 1, kb);

    // Make sure its flushed out
    rs_next_triple(r);

    if (rs_num_triples(r)) {
	ret->data.list->items[0] = okbc_new_node(OKBC_URI,
			okbc_new_atom((char *)thing));
	ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));
    } else {
	ret->data.list->items[0] = okbc_new_node(OKBC_URI, "false");
	ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(0));
    }
    rs_free_result(r);

    return ret;
}

/* okbc_get_frame_name(frame, kb, kb_local_only_p)
 *
 * (frame &key kb kb-local-only-p)
 *
 * Return Value(s): frame-name 
 * 
 * Flags: M R 
 * 
 * Returns frame-name, an entity that is the name of the frame
 * identified by frame, usually a symbol or string.
 */

okbc_node *okbc_get_frame_name(const char *frame, const char *kb,
			       int kb_local_only_p)
{
    return okbc_new_node(OKBC_URI, okbc_new_atom((char *)frame));
}

/* okbc_get_frame_pretty_name(frame, kb, kb_local_only_p)
 *
 * (frame &key kb kb-local-only-p)
 *
 * Return Value(s): string 
 * 
 * Flags: M R 
 * 
 * Returns a string that is a pretty, printed representation for frame
 * - that is, the name is suitable for use within a user interface for
 * display purposes.
 *
 * There is no guarantee that it will be possible to find a unique frame
 * given only its pretty-name, but get-frames-matching can be used to
 * find frames matching such strings when possible.
 */

okbc_node *okbc_get_frame_pretty_name(const char *frame, const char *kb,
				      int kb_local_only_p)
{
    okbc_node *ret;
    rs_result *r;
    rs_triple *t;

    r = rs_find_triples(okbc_intl_db, frame, RDFS_LABEL, NULL, ObjLiteral, 1, kb);
    t = rs_next_triple(r);
    if (t) {
        ret = okbc_new_node(OKBC_LITERAL, okbc_new_atom(t->object));
    } else {
	ret = okbc_new_node(OKBC_LITERAL, "");
    }
    rs_free_result(r);

    return ret;
}

/* okbc_get_frame_sentences(frame, kb, number_of_values, okbc_sentences_p,
 * value_selector, kb_local_only_p)
 *
 * (frame &key kb (number-of-values :all) (okbc-sentences-p true)
 * (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): list-of-sentences exact-p more-status 
 * 
 * Flags: O R 
 * 
 * Returns a list of all the logical sentences associated with a frame.  The
 * sentences may have been asserted using tell, or any other OKBC update
 * operation. If okbc-sentences-p is true, then all sentences are returned,
 * including the ones that are equivalent to basic OKBC operations. The
 * sentences equivalent to OKBC operations are defined in Table 3.3. If
 * okbc-sentences-p is false, sentences that are equivalent to OKBC operations
 * are not returned. This is very useful for user interface applications that
 * do not want to present redundant information. If no matching sentences are
 * found, list-of-sentences will be ().
 */

okbc_node *okbc_get_frame_sentences(const char *frame, const char *kb,
				    int number_of_values,
				    int okbc_sentences_p,
				    const char *value_selector,
				    int kb_local_only_p)
{
    okbc_node *ret = okbc_new_list_node(3);
    okbc_node *list;
    okbc_node *list2;
    int list_len;
    rs_result *r;
    rs_triple *t;
    int all = number_of_values > 0 ? 0 : 1;
    int i = 0;

    r = rs_find_triples(okbc_intl_db, frame, NULL, NULL, ObjAny,
		    number_of_values, kb);
    list_len = rs_num_triples(r);
    list = okbc_new_list_node(list_len);
    while ((t = rs_next_triple(r)) && (all || number_of_values-- > 0)) {
	okbc_node *statement = okbc_new_list_node(3);
	statement->data.list->items[0] =  okbc_new_node(OKBC_URI,
			okbc_new_atom(t->subject));
	statement->data.list->items[1] =  okbc_new_node(OKBC_URI,
			okbc_new_atom(t->predicate));
	if (t->literal) {
	    statement->data.list->items[2] = okbc_new_node(OKBC_LITERAL,
			    okbc_new_atom(t->object));
	} else {
	    statement->data.list->items[2] = okbc_new_node(OKBC_URI,
			    okbc_new_atom(t->object));
	}
	list->data.list->items[i++] = statement;
    }
    rs_free_result(r);

    r = rs_find_triples(okbc_intl_db, NULL, NULL, frame, ObjURI,
		    number_of_values, kb);
    list2 = okbc_new_list_node(list_len + rs_num_triples(r));
    for (i = 0; i<list_len; i++) {
	list2->data.list->items[i] = list->data.list->items[i];
    }
    free(list);
    while ((t = rs_next_triple(r)) && (all || number_of_values-- > 0)) {
	okbc_node *statement = okbc_new_list_node(3);
	statement->data.list->items[0] =  okbc_new_node(OKBC_URI,
			okbc_new_atom(t->subject));
	statement->data.list->items[1] =  okbc_new_node(OKBC_URI,
			okbc_new_atom(t->predicate));
	statement->data.list->items[2] = okbc_new_node(OKBC_URI,
			okbc_new_atom(t->object));
	list2->data.list->items[i++] = statement;
    }
    rs_free_result(r);

    ret->data.list->items[0] = list2;
    ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));
    ret->data.list->items[2] = okbc_new_node(OKBC_BOOL, ATOM(0));

    return ret;
}

/* okbc_get_frame_slots(frame, kb, inference_level, slot_type, kb_local_only_p)
 *
 * (frame &key kb (inference-level :taxonomic) (slot-type :all) kb-local-only-p)
 *
 * Return Value(s): list-of-slots exact-p 
 * 
 * Flags: E M R 
 * 
 * Returns list-of-slots, a list of all the own, template, or own
 * and template slots that are associated with frame, depending on the
 * value of slot-type.
 */

okbc_node *okbc_get_frame_slots(const char *frame, const char *kb,
				const char *inference_level,
				const char *slot_type, int kb_local_only_p)
{
    okbc_node *ret = okbc_new_list_node(2);
    okbc_node *list;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    r = rs_find_triples(okbc_intl_db, frame, NULL, NULL, ObjAny, 0, kb);
    list = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
	list->data.list->items[i++] = okbc_new_node(OKBC_URI,
			okbc_new_atom(t->predicate));
    }
    rs_free_result(r);

    ret->data.list->items[0] = list;
    ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));
    return ret;
}

/* okbc_get_frame_type(thing, kb, kb_local_only_p)
 *
 * (thing &key kb kb-local-only-p)
 *
 * Return Value(s): frame-type 
 * 
 * Flags: O R 
 * 
 * When thing identifies a frame, returns either
 * :slot, :facet, :class, or :individual,
 * depending on the type of the frame.
 * When thing does not identify a frame, frame-type is 
 * false. :Slot and :facet will be returned only in
 * those systems that support the values :slot and :facet,
 * respectively, for the :are-frames behavior.
 */

okbc_node *okbc_get_frame_type(const char *thing, const char *kb,
			       int kb_local_only_p)
{
    okbc_node *ret;
    rs_result *r;
    rs_triple *t;

    r = rs_find_triples(okbc_intl_db, thing, RDF_TYPE, NULL, ObjURI, 1, kb);
    t = rs_next_triple(r);
    if (t && !strcmp(t->object, RDFS_CLASS)) {
	ret = okbc_new_node(OKBC_LITERAL, ":class");
    } else if (t && !strcmp(t->object, RDF_PROPERTY)) {
	ret = okbc_new_node(OKBC_LITERAL, ":slot");
    } else {
	ret = okbc_new_node(OKBC_LITERAL, ":individual");
    }
    rs_free_result(r);

    return ret;
}

/* okbc_get_frames_matching(pattern, kb, wildcards_allowed_p, selector,
 * force_case_insensitive_p, kb_local_only_p)
 *
 * (pattern &key kb (wildcards-allowed-p true) (selector :all)
 * force-case-insensitive-p kb-local-only-p)
 *
 * Return Value(s): matching-frames longest-matching-substring 
 * 
 * Flags: E O R 
 * 
 * Given a pattern, which is a string or a symbol, finds a set of matching
 * frames for that pattern.  The match of a frame to a pattern could take into
 * account the frame's name (if meaningful), printed representation,
 * pretty-name, or any KB-specific feature such as a list of synonyms.
 *
 * Returns the following two values:
 *
 * 1. matching-frames - The list of matching frames (which is () if no matches
 * are found).
 *
 * 2. longest-matching-substring - The longest matching initial substring. This
 * returned value is useful in applications that use get-frames-matching to
 * implement a completion facility, or prompt users for frames (false if no
 * matches are found).  Wildcards-allowed-p
 *
 * -- When true, the pattern may contain * (zero or more characters) and ?
 * (exactly one character) wildcards. Wildcard characters are escaped with the
 * backslash character. If this argument is false, the * and ? characters
 * simply denote themselves and need not be escaped. Selector
 *
 * -- May be a procedure (see Section 5) of signature (candidate-name, kb,
 * kb-local-only-p) that returns true if the candidate name is to be accepted
 * and false otherwise, or one of the following keywords:
 * 
 * :all - Select all frames :class - Select only class frames :individual -
 * Select only individual frames :slot - Select only slot frames :facet -
 * Select only facet frames Force-Case-Insensitive-P -- When true, cause the
 * comparison is to be case-insensitive, irrespective of the IO syntax of the
 * KB.
 */

okbc_node *okbc_get_frames_matching(const char *pattern, const char *kb,
				    int wildcards_allowed_p,
				    const char *selector,
				    int force_case_insensitive_p,
				    int kb_local_only_p)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_frames_with_facet_value(slot, facet, value, kb, inference_level, slot_type, value_selector, kb_local_only_p)
 *
 * (slot facet value &key kb (inference-level :taxonomic) (slot-type :own) (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): frames exact-p 
 * 
 * Flags: O R 
 * 
 * Returns the set of frames in which the specified facet value is accessible
 * on the specified slot.
 * If the system is unable to find any frame/slot/facet combinations with the
 * specified value, () is returned.
 * This operation allows user interfaces to take users from a value
 * displayed as a facet value on a particular frame/slot to the place
 * that asserted the value.
 */

okbc_node *okbc_get_frames_with_facet_value(const char *slot,
					    const char *facet,
					    const char *value,
					    const char *kb,
					    const char *inference_level,
					    const char *slot_type,
					    const char *value_selector,
					    int kb_local_only_p)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_frames_with_slot_value(slot, value, kb, inference_level, slot_type,
 * value_selector, kb_local_only_p)
 *
 * (slot value &key kb (inference-level :taxonomic) (slot-type :own)
 * (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): frames exact-p 
 * 
 * Flags: O R 
 * 
 * Returns the set of frames in which the specified slot value is accessible.
 * If the system is unable to find any frame/slot combinations with the
 * specified value, () is returned.  This operation allows user interfaces to
 * take users from a value displayed as a slot value on a particular frame to
 * the place that asserted the value.
 */

okbc_node *okbc_get_frames_with_slot_value(const char *slot,
					   const char *value,
					   const char *kb,
					   const char *inference_level,
					   const char *slot_type,
					   const char *value_selector,
					   int kb_local_only_p)
{
    okbc_node *ret;
    okbc_node *list;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    ret = okbc_new_list_node(2);
    r = rs_find_triples(okbc_intl_db, NULL, slot, value, ObjAny, 0, kb);
    list = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
	if (t->literal) {
	    list->data.list->items[i++] =
			okbc_new_node(OKBC_LITERAL, okbc_new_atom(t->subject));
	} else {
	    list->data.list->items[i++] =
			okbc_new_node(OKBC_URI, okbc_new_atom(t->subject));
	}
    }
    rs_free_result(r);
    ret->data.list->items[0] = list;
    ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));

    return ret;
}

/* okbc_get_instance_types(frame, kb, inference_level, number_of_values,
 * kb_local_only_p)
 *
 * (frame &key kb (inference-level :taxonomic) (number-of-values :all)
 * kb-local-only-p)
 *
 * Return Value(s): list-of-types exact-p more-status 
 * 
 * Flags: E M R 
 * 
 * Returns the list-of-types of frame, that is, the list of 
 * classes of which frame is an instance.
 */

okbc_node *okbc_get_instance_types(const char *frame, const char *kb,
				   const char *inference_level,
				   int number_of_values,
				   int kb_local_only_p)
{
    okbc_node *ret = NULL;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    if (!inference_level || inference_level == key_taxonomic) {
	r = rs_find_triples(okbc_intl_db, frame, RDF_TYPE, NULL, ObjURI,
			number_of_values, kb);
    } else {
	r = rs_find_asserted_triples(okbc_intl_db, frame, RDF_TYPE, NULL,
				     ObjURI, number_of_values, kb);
    }
    ret = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
        ret->data.list->items[i++] = okbc_new_node(OKBC_URI,
			okbc_new_atom(t->object));
    }
    rs_free_result(r);

    return ret;
}

/* okbc_get_kb_behaviors(kb_type_or_kb, connection)
 *
 * (&key (kb-type-or-kb (current-kb)) (connection (local-connection)))
 *
 * Return Value(s): list-of-behaviors 
 * 
 * Flags: M R 
 * 
 * When kb-type-or-kb is either a KB or a kb-type, returns
 * list-of-behaviors, which is a list of keywords naming all
 * the behaviors recognized by this KB, or identified by the kb-type,
 * respectively.
 */

okbc_node *okbc_get_kb_behaviors(const char *kb_type_or_kb,
				 const char *connection)
{
    okbc_node *ret = okbc_new_list_node(0);

    return ret;
}

/* okbc_get_kb_classes(kb, selector, kb_local_only_p)
 *
 * (&key kb (selector :system-default) kb-local-only-p)
 *
 * Return Value(s): list-of-classes 
 * 
 * Flags: E O R 
 * 
 * Returns list-of-classes, a list of the classes in the KB.
 * Selector can be one of the following:
 * 
 * :all - Returns all classes
 * :frames - Returns classes that are represented as frames
 * :system-default - Returns either all classes or
 * only class frames, according to which is the KRS's default
 */

okbc_node *okbc_get_kb_classes(const char *kb, const char *selector,
			       int kb_local_only_p)
{
    okbc_node *ret = NULL;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    r = rs_find_triples(okbc_intl_db, NULL, RDF_TYPE, RDFS_CLASS, ObjURI, 0, kb);
    ret = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
	ret->data.list->items[i++] = okbc_new_node(OKBC_URI, okbc_new_atom(t->subject));
    }

    return ret;
}

/* okbc_get_kb_direct_children(kb)
 *
 * (&key kb)
 *
 * Return Value(s): list-of-child-kbs 
 * 
 * Flags: E O R 
 * 
 * Returns the list-of-child-kbs - that is, the list of KBs that directly
 * include kb. Note that certain KB implementations may allow circular
 * inclusion dependencies in KBs. The semantics of KB inclusion are not
 * specified by OKBC, but where possible, processing can be limited to a
 * particular KB by the use of the kb-local-only-p argument.
 */

okbc_node *okbc_get_kb_direct_children(const char *kb)
{
    okbc_node *ret = okbc_new_list_node(0);

    return ret;
}

/* okbc_get_kb_direct_parents(kb)
 *
 * (&key kb)
 *
 * Return Value(s): list-of-parent-kbs 
 * 
 * Flags: E O R 
 * 
 * Returns the list-of-parent-kbs - that is, the list of KBs directly included
 * by kb. Note that certain KB implementations may allow circular inclusion
 * dependencies in KBs. The semantics of KB inclusion are not specified by
 * OKBC, but where possible, processing can be limited to a particular KB by
 * the use of the kb-local-only-p argument.
 */

okbc_node *okbc_get_kb_direct_parents(const char *kb)
{
    okbc_node *ret = okbc_new_list_node(0);

    return ret;
}

/* okbc_get_kb_facets(kb, selector, kb_local_only_p)
 *
 * (&key kb (selector :system-default) kb-local-only-p)
 *
 * Return Value(s): list-of-facets 
 * 
 * Flags: E O R 
 * 
 * Returns the list-of-facets in kb.
 * Selector can be one of the following:
 * 
 * :all - Returns all facets
 * :frames - Returns facets that are represented as frames
 * :system-default - Returns either all facets or
 * only facets represented as frames, according to which is the KRS's
 * default
 */

okbc_node *okbc_get_kb_facets(const char *kb, const char *selector,
			      int kb_local_only_p)
{
    okbc_node *ret = okbc_new_list_node(0);

    return ret;
}

/* okbc_get_kb_frames(kb, kb_local_only_p)
 *
 * (&key kb kb-local-only-p)
 *
 * Return Value(s): list-of-frames 
 * 
 * Flags: E M R 
 * 
 * Returns the list-of-frames in the KB, including class, slot,
 * facets and individual frames, when present.
 */

okbc_node *okbc_get_kb_frames(const char *kb, int kb_local_only_p)
{
    okbc_node *ret;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    r = rs_find_all_resources(okbc_intl_db, 0, kb);
    ret = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
	ret->data.list->items[i++] = okbc_new_node(OKBC_URI,
			okbc_new_atom(t->subject));
    }
    rs_free_result(r);

    return ret;
}

/* okbc_get_kb_individuals(kb, selector, kb_local_only_p)
 *
 * (&key kb (selector :system-default) kb-local-only-p)
 *
 * Return Value(s): list-of-individuals 
 * 
 * Flags: E O R 
 * 
 * Returns list-of-individuals, a list of the individual frames in
 * kb. Selector can be one of the following:
 * 
 * :all - Returns all accessible individuals
 * :frames - Returns only individuals that are frames
 * :system-default - Returns either all individuals or
 * only individual frames, according to which is the KRS's default
 */

okbc_node *okbc_get_kb_individuals(const char *kb, const char *selector,
				   int kb_local_only_p)
{
    // XXX Needs implementing
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_kb_roots(kb, selector, kb_local_only_p)
 *
 * (&key kb (selector :all) kb-local-only-p)
 *
 * Return Value(s): list-of-roots 
 * 
 * Flags: E O R 
 * 
 * Every KB has one or more frames at the top (root) of the
 * KB. A frame C is a root of the KB K if there exists no class D such
 * that D is a superclass of C and D is in the KB K and if there
 * exists no class E such that E is a type of C and E is in the
 * KB K, or available in K when kb-local-only-p is false. This
 * operation identifies and returns those roots, the list-of-roots.
 * Note that this means that unparented individuals, slots and facets will
 * also be returned.
 *
 * Some KRSs allow user-defined classes to be roots of a KB, whereas
 * other KRSs always import certain system-defined classes (for example,
 * thing) into each KB and force all
 * user classes to be subclasses of thing. These system-defined classes
 * may normally be invisible to the user in some KRSs. The
 * selector argument controls which root classes are returned as
 * follows:
 *
 * selector = :all returns all the true root classes of the 
 * KB, regardless of whether they are system-defined or 
 * user-defined.
 * selector = :user returns the user-defined
 * root classes of the KB, namely all classes C available in the KB such that
 * C was defined by a user application as opposed to being a built-in
 * part of every KB, and such that there exists no class D that is both
 * user-defined and a superclass of C. That is, there may
 * exist system-defined superclasses of C.
 * 
 * If kb-local-only-p is true, the list returned may return only
 * the root classes defined in kb itself; classes that were
 * inherited from other (included) KBs may be excluded. This means
 * that a class that has superclasses in some KB included by kb,
 * but has no superclasses defined in kb, may be returned as a
 * root class if kb-local-only-p is true.
 *
 */

okbc_node *okbc_get_kb_roots(const char *kb, const char *selector,
			     int kb_local_only_p)
{
    okbc_node *ret = okbc_new_list_node(1);

    ret->data.list->items[0] = okbc_new_node(OKBC_URI,
		    okbc_new_atom(RDFS_RESOURCE));

    return ret;
}

/* okbc_get_kb_slots(kb, selector, kb_local_only_p)
 *
 * (&key kb (selector :system-default) kb-local-only-p)
 *
 * Return Value(s): list-of-slots 
 * 
 * Flags: E O R 
 * 
 * Returns the list-of-slots that are defined in the KB.
 * Selector can be one of the following:
 * 
 * :all - Returns all slots
 * :frames - Returns slots that are represented frames
 * :system-default - Returns either all slots or
 * only slot frames, according to which is the KRS's default
 */

okbc_node *okbc_get_kb_slots(const char *kb, const char *selector,
			     int kb_local_only_p)
{
    okbc_node *ret = NULL;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    r = rs_find_triples(okbc_intl_db, NULL, RDF_TYPE, RDF_PROPERTY, ObjURI, 0,
		    kb);
    ret = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
	ret->data.list->items[i++] = okbc_new_node(OKBC_URI,
			okbc_new_atom(t->subject));
    }
    rs_free_result(r);

    return ret;
}

/* okbc_get_kb_type(thing, connection)
 *
 * (thing &key (connection (local-connection)))
 *
 * Return Value(s): kb-type 
 * 
 * Flags: O R 
 * 
 * Returns the kb-type object for thing, which may be a KB,
 * a kb-locator, or a kb-type name. KB-type names are KRS-specific, and
 * will be documented with the KRS being used. KB-type names need never be
 * used, since a kb-type can be selected by a user or application by using
 * the get-kb-types and frs-name operations. It is not
 * portable to specify a kb-type name as a literal in an OKBC program.
 */

okbc_node *okbc_get_kb_type(const char *thing, const char *connection)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_kb_types(connection)
 *
 * (&key (connection (local-connection)))
 *
 * Return Value(s): list-of-kb-types 
 * 
 * Flags: E O R 
 * 
 * Returns a list of KB types for each of the known KRSs accessible
 * through the connection. List-of-kb-types contains one
 * kb-type entry for each KRS known through the connection, and possibly also
 * kb-type objects representing supertypes of the KRSs supported.
 */

okbc_node *okbc_get_kb_types(const char *connection)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_kbs(connection)
 *
 * (&key (connection (local-connection)))
 *
 * Return Value(s): list-of-kbs 
 * 
 * Flags: E O R 
 * 
 * Returns a list-of-kbs containing all the known KBs accessible 
 * through the connection, irrespective of the KB's implementation
 * type. Note: the connection from which each of the KBs returned was derived
 * is not necessarily the same as the value of the connection
 * argument. Later processing of any of these KBs should be done with
 * respect to the connection returned when the connection operation
 * is applied to the KB.
 */

okbc_node *okbc_get_kbs(const char *connection)
{
    okbc_node *ret = NULL;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    r = rs_find_all_models(okbc_intl_db);
    ret = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
	ret->data.list->items[i++] = okbc_new_node(OKBC_URI, okbc_new_atom(t->subject));
    }
    rs_free_result(r);

    return ret;
}

/* okbc_get_kbs_of_type(kb_type, connection)
 *
 * (&key kb-type (connection (local-connection)))
 *
 * Return Value(s): list-of-kbs 
 * 
 * Flags: E M R 
 * 
 * Returns list-of-kbs, the list of all the known KBs whose type
 * matches kb-type, and that are accessible through the
 * connection.
 */

okbc_node *okbc_get_kbs_of_type(const char *kb_type,
				const char *connection)
{
    return okbc_get_kbs(connection);
}

/* okbc_get_procedure(name, kb)
 *
 * (name &key kb)
 *
 * Return Value(s): procedure 
 * 
 * Flags: O R 
 * 
 * Returns the procedure that is the procedure association for
 * the name, or false if there is no such procedure association.
 * See register-procedure, unregister-procedure, and
 * call-procedure.
 */

okbc_node *okbc_get_procedure(const char *name, const char *kb)
{
    // XXX Needs implementing when we have proper stored procedures
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_slot_facets(frame, slot, kb, inference_level, slot_type, kb_local_only_p)
 *
 * (frame slot &key kb (inference-level :taxonomic) (slot-type :own) kb-local-only-p)
 *
 * Return Value(s): list-of-facets exact-p 
 * 
 * Flags: E M R 
 * 
 * Returns the list-of-facets associated with slot in
 * frame.
 */

okbc_node *okbc_get_slot_facets(const char *frame, const char *slot,
				const char *kb,
				const char *inference_level,
				const char *slot_type, int kb_local_only_p)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_slot_type(frame, slot, kb, inference_level, kb_local_only_p)
 *
 * (frame slot &key kb (inference-level :taxonomic) kb-local-only-p)
 *
 * Return Value(s): slot-type 
 * 
 * Flags: O R 
 * 
 * Returns one of {:own, :template, false} to
 * identify the slot-type of the slot on question. If there are both
 * an own and a template slot on frame identified by slot, then
 * :own is returned.
 * If no such slot is known, then false is returned.
 */

okbc_node *okbc_get_slot_type(const char *frame, const char *slot,
			      const char *kb, const char *inference_level,
			      int kb_local_only_p)
{
    okbc_node *ret = NULL;

    return ret;
}

/* okbc_get_slot_value(frame, slot, kb, inference_level, slot_type, value_selector, kb_local_only_p)
 *
 * (frame slot &key kb (inference-level :taxonomic) (slot-type :own) (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): value-or-false exact-p 
 * 
 * Flags: O R 
 * 
 * Returns the single member of the set of values
 * of the slot. This operation is meaningful only for single-valued
 * slots. It is an error to call get-slot-value on a non-single-valued 
 * slot, and implementations should signal a cardinality-violation if 
 * this occurs. When there is no value for the slot, value-or-false 
 * is false.
 */

okbc_node *okbc_get_slot_value(const char *frame, const char *slot,
			       const char *kb, const char *inference_level,
			       const char *slot_type,
			       const char *value_selector,
			       int kb_local_only_p)
{
    okbc_node *ret;
    rs_result *r;
    rs_triple *t;

    r = rs_find_triples(okbc_intl_db, frame, slot, NULL, ObjAny, 1, kb);
    ret = okbc_new_list_node(rs_num_triples(r));
    t = rs_next_triple(r);
    if (t) {
        if (t->literal) {
            ret->data.list->items[0] =
		        okbc_new_node(OKBC_LITERAL, okbc_new_atom(t->object));
        } else {
            ret->data.list->items[0] =
		        okbc_new_node(OKBC_URI, okbc_new_atom(t->object));
        }
    } else {
	ret->data.list->items[0] = okbc_new_node(OKBC_LITERAL, "false");
    }
    rs_free_result(r);
    ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));

    return ret;
}

/* okbc_get_slot_values(frame, slot, kb, inference_level, slot_type,
 * number_of_values, value_selector, kb_local_only_p)
 *
 * (frame slot &key kb (inference-level :taxonomic) (slot-type :own)
 * (number-of-values :all) (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): list-of-values exact-p more-status 
 * 
 * Flags: E O R 
 * 
 * Returns the list-of-values of slot within frame. If the :collection-type of
 * the slot is :list, and only :direct own slots have been asserted, then order
 * is preserved; otherwise, the values are returned in no guaranteed order.
 * Get-slot-values always returns a list of values. If the specified slot has
 * no values, () is returned.
 */

okbc_node *okbc_get_slot_values(const char *frame, const char *slot,
				const char *kb,
				const char *inference_level,
				const char *slot_type,
				int number_of_values,
				const char *value_selector,
				int kb_local_only_p)
{
    okbc_node *ret;
    okbc_node *list;
    rs_result *r;
    rs_triple *t;
    int i = 0;

    ret = okbc_new_list_node(3);
    r = rs_find_triples(okbc_intl_db, frame, slot, NULL, ObjAny, 0, kb);
    list = okbc_new_list_node(rs_num_triples(r));
    while ((t = rs_next_triple(r))) {
	if (t->literal) {
	    list->data.list->items[i++] =
			okbc_new_node(OKBC_LITERAL, okbc_new_atom(t->object));
	} else {
	    list->data.list->items[i++] =
			okbc_new_node(OKBC_URI, okbc_new_atom(t->object));
	}
    }
    rs_free_result(r);
    ret->data.list->items[0] = list;
    ret->data.list->items[1] = okbc_new_node(OKBC_BOOL, ATOM(1));
    ret->data.list->items[2] = okbc_new_node(OKBC_BOOL, ATOM(0));

    return ret;
}

/* okbc_get_slot_values_in_detail(frame, slot, kb, inference_level, slot_type,
 * number_of_values, value_selector, kb_local_only_p)
 *
 * (frame slot &key kb (inference-level :taxonomic) (slot-type :own)
 * (number-of-values :all) (value-selector :either) kb-local-only-p)
 *
 * Return Value(s): list-of-specs exact-p more-status default-p 
 * 
 * Flags: E M R 
 * 
 * Returns the list-of-specs describing the values of slot within frame. If the
 * :collection-type of the slot is :list, and only :direct own slots have been
 * asserted, then order is preserved; otherwise, the values are returned in no
 * guaranteed order. Get-slot-values-in-detail always returns a list of
 * specifications as its list-of-specs value. If the specified slot has no
 * values, () is returned.
 *
 * Each spec is a 3-tuple of the form (value direct-p default-p).
 * 
 * value - A value of the slot direct-p - A flag that is true if the value is
 * known to be directly asserted for the slot and false otherwise default-p - A
 * flag that is true if the value is known to be a default value, and false
 * otherwise The default-p returned value is true if the list-of-specs is (),
 * and the fact that there are no values is itself a default.
 */

okbc_node *okbc_get_slot_values_in_detail(const char *frame,
					  const char *slot, const char *kb,
					  const char *inference_level,
					  const char *slot_type,
					  int number_of_values,
					  const char *value_selector,
					  int kb_local_only_p)
{
    okbc_node *ret = NULL;

    return ret;
}
