/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: cache.c,v 1.19 2005/01/24 16:05:45 theno23 Exp $
 */

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <mysql.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>

#include "config.h"
#include "rdfsql/rdfsql.h"

#define rs_mat_hash(m, a, b) ((a * (m->size + 1) + b) % m->hash_size)

#define TMP_S_TABLE \
"CREATE TEMPORARY TABLE `tmp_"TBLPREFIX"resources` ( \
  `a` bigint(20) NOT NULL default '0' \
)"

#define TMP_TABLE \
"CREATE TEMPORARY TABLE `tmp_"TBLPREFIX"triples` ( \
  `a` bigint(20) NOT NULL default '0', \
  `b` bigint(20) NOT NULL default '0', \
  PRIMARY KEY(a, b) \
)"

#define TMP_MSO_TABLE \
"CREATE TEMPORARY TABLE `tmp_mso` ( \
  `m` bigint(20) NOT NULL default '0', \
  `s` bigint(20) NOT NULL default '0', \
  `o` bigint(20) NOT NULL default '0', \
  PRIMARY KEY(m, s, o) \
)"

/* Number of triples accumulated before making an assertion */
#define BLOCK_SIZE 256

/* Sparese matrix representation structures */

/* List of all the [a,b]'s for which the value is true */
typedef struct _fl {
    int a;
    int b;
    struct _fl *next;
    struct _fl *next_hash;
} rs_full_list;

/* List of all the b's for this a for which the value is true */
typedef struct _sl {
    int b;
    struct _sl *next;
} rs_split_list;

/* Overall struct representing the matrix */
typedef struct {
    rs_full_list *l_current;
    rs_full_list *l_pending;
    rs_split_list **s_current;
    rs_split_list **s_pending;
    rs_full_list **hash;
    int size;
    int hash_size;
} rs_sparse_mat;

/* Matrix handling functions */

/*create new matrix */
rs_sparse_mat *rs_mat_new(int size);
/* set [a,b] to true */
int rs_mat_set(rs_sparse_mat *m, int a, int b);

gint hash_cmp(gconstpointer a, gconstpointer b);

gboolean rs_hash_equal(gconstpointer a, gconstpointer b);
guint rs_hash_hash(gconstpointer v);

static void rs_rdf1(RDFSQL *db);
static void rs_rdfs23(RDFSQL *db, int rule, int by_model, guint64 smodel);
static void rs_rdfs4(RDFSQL *db);

static void rs_block_append(RDFSQL *db, rs_raw_triple *block, int *count,
                            guint64 s, guint64 p, guint64 o);
static void rs_block_flush(RDFSQL *db, rs_raw_triple *block, int *count);

/* rebuild_latice_inline:
 *
 * latice rebuilting code. This produces flagged triples in the main triples
 * table.
 *
 * type: the rdf:type of the individulas that are the nodes on the latice.
 * rel: the relation that makes up the edges of the latice
 * top: the latice top, or NULL if it is free
 * bottom: the lattice bottom, or NULL if it is free
 */
static void rebuild_latice_inline(RDFSQL * db, char *type, char *rel,
				  char *top, char *bottom, int verbosity);

void rs_rebuild_taxonomy(RDFSQL * db, int verbosity)
{
    /* Mark trasitive closure cache as up-to-date */
    if (mysql_query(db->db, "UPDATE "TBLPREFIX"cache_state SET taxonomy=0")) {
	fprintf(stderr, "SQL error in cache state update %d: %s\n",
		mysql_errno(db->db), mysql_error(db->db));
    }

    /* Caclulate RDF1 and RDFS4 a+b */
    if (verbosity > 0) printf("finding entailments for rdf1\n");
    rs_rdf1(db);
    if (verbosity > 0) printf("finding entailments for rdfs4\n");
    rs_rdfs4(db);

    if (verbosity > 0) printf("finding entailments for rdfs5\n");
    /* Cacluate transitive closure of rdfs:subPropertyOf relation */
    rebuild_latice_inline(db, RDF_PROPERTY, RDFS_SUBPROPERTY, NULL, NULL,
			  verbosity);

    if (verbosity > 0) printf("finding entailments for rdfs7+8\n");
    /* Cacluate transitive closure of rdf:subClassOf relation */
    rebuild_latice_inline(db, RDFS_CLASS, RDFS_SUBCLASS, RDFS_RESOURCE, NULL,
			  verbosity);

    /* Caclulate RDFS2+3, has to be done last */
    if (verbosity > 0) printf("finding entailments for rdfs2\n");
    rs_rdfs23(db, 2, 0, 0);
    if (verbosity > 0) printf("finding entailments for rdfs3\n");
    rs_rdfs23(db, 3, 0, 0);
}

static void rebuild_latice_inline(RDFSQL *db, char *type, char *rel,
				  char *top, char *bottom, int verbosity)
{
    rs_result *res;
    rs_triple *t;
    MYSQL_RES *mres;
    MYSQL_ROW mrow;
    /* A list of the individuals in the latice */
    GSList *indivs = NULL;
    GSList *tmp = NULL, *it;
    int num_indivs;
    unsigned int i;
    rs_sparse_mat *mat;
    GHashTable *uri_lookup;
    guint64 old_model, *skey, *okey;
    guint64 rel_key, top_key, bottom_key;
    guint64 *rc_hash;
    rs_raw_triple raw_block[BLOCK_SIZE];
    int block_count = 0;
    char *this_model = g_strdup_printf("3store:cache#%s", rel);
    int pass, count;
    rs_full_list *it_f;
    rs_split_list *it_s;

    /* Find the individuals that are expliticly of the right type */
    res = rs_find_triples(db, NULL, RDF_TYPE, type, ObjURI, 0, NULL);
    while ((t = rs_next_triple(res))) {
	skey = malloc(sizeof(guint64));
	*skey = rs_hash(t->subject);
	indivs = g_slist_prepend(indivs, skey);
    }
    rs_free_result(res);

    /* Find the individuals that are implied to be of the right type
     * becuase of the relation */
    res = rs_find_triples(db, NULL, rel, NULL, ObjURI, 0, NULL);
    while ((t = rs_next_triple(res))) {
	//printf("%s\n", t->subject);
	//printf("%s\n", t->object);
	if (!t->subject) {
	    fprintf(stderr, "ERROR: Invalid subject key\n");
	    rs_free_result(res);
	    continue;
	}
	if (!t->object) {
	    fprintf(stderr, "ERROR: Invalid object key\n");
	    rs_free_result(res);
	    continue;
	}

	skey = malloc(sizeof(guint64));
	*skey = rs_hash(t->subject);
	okey = malloc(sizeof(guint64));
	*okey = rs_hash(t->object);
	
	indivs = g_slist_prepend(indivs, skey);
	indivs = g_slist_prepend(indivs, okey);
    }

    /* Find the individuals that are implied to be of the right type
     * becuase of rdfs2+3 */
    if (!strcmp(type, RDFS_CLASS)) {
	if (mysql_query(db->db, "SELECT DISTINCT object FROM "TBLPREFIX"triples WHERE predicate=-4085280037482734459")) {
	    fprintf(stderr, "Failed to do query rs23 (%d): %s\n", mysql_errno(db->db), mysql_error(db->db));
	}
	mres = mysql_store_result(db->db);
	while ((mrow = mysql_fetch_row(mres))) {
	    okey = malloc(sizeof(guint64));
	    *okey = strtoll(mrow[0], NULL, 10);
	    indivs = g_slist_prepend(indivs, okey);
	}
	mysql_free_result(mres);
    }

    rs_free_result(res);

    /* sort+uniq the list to remove duplicates */
    indivs = g_slist_sort(indivs, hash_cmp);

    for (it = indivs; it; it = it->next) {
	if (!it->next || *(guint64 *)(it->data) != *(guint64 *)(it->next->data)) {
	    tmp = g_slist_prepend(tmp, it->data);
	}
    }
    indivs = tmp;
    num_indivs = g_slist_length(indivs);

    rc_hash = malloc(num_indivs * sizeof(guint64));
    uri_lookup = g_hash_table_new(rs_hash_hash, rs_hash_equal);
    for (i = 0, it = tmp; it; it = it->next, i++) {
	g_hash_table_insert(uri_lookup, it->data, (gpointer) i);
	rc_hash[i] = *(guint64 *)(it->data);
	//printf("%s\n", (char *)it->data);
    }

    /* Nasty hack to make sure we dont wipe out the callers data */
    old_model = db->model;
    db->model = rs_check_model(db, this_model);
    rs_flush(db);
    free(this_model);

    rel_key = rs_hash(rel);
    top_key = rs_hash(top);
    bottom_key = rs_hash(bottom);

    if (mysql_query(db->db, "LOCK TABLES "TBLPREFIX"triples WRITE, "TBLPREFIX"resources WRITE, "TBLPREFIX"literals WRITE, "TBLPREFIX"cache_state WRITE")) {
	fprintf(stderr, "Failed to lock tables (%d): %s\n", mysql_errno(db->db), mysql_error(db->db));
    }

    for (it = indivs; it; it=it->next) {
	guint64 indiv = *(guint64 *)(it->data);

	/* Assert that a foo is a subfoo of itsself */
	rs_block_append(db, raw_block, &block_count,
			indiv, rel_key, indiv);
	if (top) {
	    /* Assert that a foo is a subfoo of top */
	    rs_block_append(db, raw_block, &block_count,
			indiv, rel_key, top_key);
	}
	if (bottom) {
	    /* Assert that a foo is a superfoo of bottom */
	    rs_block_append(db, raw_block, &block_count,
			bottom_key, rel_key, indiv);
	}
    }
    rs_block_flush(db, raw_block, &block_count);

    /* The Floyd-Warshall's algorithm:
     *
     * adjacent[X][Z] and adjacent[Z][Y] => adjacent[X][Y]
     */

    mat = rs_mat_new(num_indivs);

    if (mysql_query(db->db, "UNLOCK TABLES")) {
	fprintf(stderr, "Failed to unlock tables (%d): %s\n", mysql_errno(db->db), mysql_error(db->db));
    }

    /* Find ground facts about relation, and populate the matrix with them */
    res = rs_find_triples(db, NULL, rel, NULL, ObjURI, 0, NULL);

    /*
    if (mysql_query(db->db, "LOCK TABLES "TBLPREFIX"triples WRITE, "TBLPREFIX"resources WRITE, "TBLPREFIX"literals WRITE, "TBLPREFIX"cache_state WRITE")) {
	fprintf(stderr, "Failed to lock tables (%d): %s\n", mysql_errno(db->db), mysql_error(db->db));
    }
    */

    while ((t = rs_next_triple(res))) {
	int c, r;
	guint64 h;

	if (!(t->subject) || !(t->object)) {
	    fprintf(stderr, "Skipping invalid key\n");
	    continue;
	}
	/* The sub-foo is the matrix column */
	h = rs_hash(t->subject);
	c = (int) g_hash_table_lookup(uri_lookup, &h);
	/* And the super-foo is the row */
	h = rs_hash(t->object);
	r = (int) g_hash_table_lookup(uri_lookup, &h);

	rs_mat_set(mat, c, r);
    }

    pass = 0;
    while (mat->l_pending) {
        /* roll pending and current matricies */
        mat->l_current = mat->l_pending;
        mat->l_pending = NULL;

        for (i=0; i<mat->size; i++) {
            mat->s_current[i] = mat->s_pending[i];
            mat->s_pending[i] = NULL;
        }

        count = 0;
        for (it_f = mat->l_current; it_f; it_f = it_f->next) {
            for (it_s = mat->s_current[it_f->b]; it_s; it_s = it_s->next) {
                if (rs_mat_set(mat, it_f->a, it_s->b)) {
                    /* assert */
		    rs_block_append(db, raw_block, &block_count,
				    rc_hash[it_f->a], rel_key,
                                    rc_hash[it_s->b]);
                    count++;
                }
            }
        }

        for (it_f = mat->l_current; it_f; it_f = it_f->next) {
            for (it_s = mat->s_pending[it_f->b]; it_s; it_s = it_s->next) {
                if (rs_mat_set(mat, it_f->a, it_s->b)) {
                    /* assert */
		    rs_block_append(db, raw_block, &block_count,
				    rc_hash[it_f->a], rel_key,
                                    rc_hash[it_s->b]);
                    count++;
                }
            }
        }
	if (verbosity > 1) {
	    printf("added %d\n", count);
	}
    }
    rs_block_flush(db, raw_block, &block_count);
/*
    if (mysql_query(db->db, "UNLOCK TABLES")) {
	fprintf(stderr, "Failed to unlock tables (%d): %s\n", mysql_errno(db->db), mysql_error(db->db));
    }
    */

    /* Restore the callers model hash */
    db->model = old_model;
}

gint hash_cmp(gconstpointer a, gconstpointer b)
{
    guint64 av, bv;

    av = *(guint64 *)a;
    bv = *(guint64 *)b;

    if (av == bv) {
	return 0;
    } else if (av > bv) {
	return 1;
    }

    return -1;
}

rs_sparse_mat *rs_mat_new(int size)
{
    rs_sparse_mat *m;

    m = malloc(sizeof(rs_sparse_mat));
    m->size = size;
    m->hash_size = size * 5;
    m->l_current = NULL;
    m->l_pending = NULL;
    m->s_current = calloc(size, sizeof(rs_split_list));
    m->s_pending = calloc(size, sizeof(rs_split_list));
    m->hash = calloc(m->hash_size, sizeof(rs_full_list));

    return m;
}

int rs_mat_set(rs_sparse_mat *m, int a, int b)
{
    rs_split_list *it;
    rs_full_list *ni, *hash_pt;

    hash_pt = m->hash[rs_mat_hash(m, a, b)];
    for (ni = hash_pt; ni; ni=ni->next_hash) {
        if (ni->a == a && ni->b == b) {
            /* Its allready set */
            return 0;
        }
    }

    /* Its not been set allready, set it */
    it = malloc(sizeof(rs_split_list));
    it->b = b;
    it->next = m->s_pending[a];
    m->s_pending[a] = it;

    ni = malloc(sizeof(rs_full_list));
    ni->a = a;
    ni->b = b;
    ni->next = m->l_pending;
    m->l_pending = ni;

    /* The hash shares data with the list (to save bytes), so we just needs to
     * set its 'next' ptr */
    ni->next_hash = hash_pt;
    m->hash[rs_mat_hash(m, a, b)] = ni;

    return 1;
}

gboolean rs_hash_equal(gconstpointer a, gconstpointer b)
{
    return (*(guint64 *)a == *(guint64 *)b);
}

guint rs_hash_hash(gconstpointer v)
{
    return *(guint64 *)v;
}

static void rs_block_append(RDFSQL *db, rs_raw_triple *block, int *count,
                            guint64 s, guint64 p, guint64 o)
{
    block[*count].subj = s;
    block[*count].pred = p;
    block[*count].obj = o;
    if (++*count == BLOCK_SIZE) {
	rs_block_flush(db, block, count);
    }
}

static void rs_block_flush(RDFSQL *db, rs_raw_triple *block, int *count)
{
    if (*count > 0) {
	if (rs_assert_raw_block(db, block, *count, ObjURI, 1)) {
	    fprintf(stderr, "Fatal error, exiting\n");
	    exit(1);
	}
    }
    *count = 0;
}

/* rs_rdf1: cacluate and assert RDF1 entailments */
static void rs_rdf1(RDFSQL *db)
{
    guint64 model, type, property;
    char *query;

    model = rs_check_model(db, "3store:cache#RDF1");
    rs_flush_other(db, model);

    type = rs_hash(RDF_TYPE);
    property = rs_hash(RDF_PROPERTY);
   
    if (mysql_query(db->db, TMP_S_TABLE)) {
	fprintf(stderr, "Failed to create tmp table (%d): %s\n", mysql_errno(db->db), mysql_error(db->db));
    }

    if (mysql_query(db->db, "INSERT INTO tmp_"TBLPREFIX"resources SELECT DISTINCT predicate FROM "TBLPREFIX"triples")) {
	fprintf(stderr, "Failed to find all predicates (%d): %s\n", mysql_errno(db->db), mysql_error(db->db));
	return;
    }

    query = g_strdup_printf("INSERT INTO "TBLPREFIX"triples SELECT %lld, a, %lld, %lld, 0, 1 FROM tmp_"TBLPREFIX"resources", model, type, property);
    if (mysql_query(db->db, query)) {
	fprintf(stderr, "Failed to transfer results '%s' (%d): %s\n", query, mysql_errno(db->db), mysql_error(db->db));
    }
    free(query);

    mysql_query(db->db, "DROP TABLE tmp_"TBLPREFIX"resources");
}

/* rs_rdfs23: calculate and assert RDFS2 or RDFS3 entailments, if by_model is
 * set only triples from smodel will be considered for the typing constrinat,
 * though all will be used for the range and domain calculations. Rule
 * specifies wether RDFS 2 or 3 should be used */

static void rs_rdfs23(RDFSQL *db, int rule, int by_model, guint64 smodel)
{
    guint64 predicate = 0, type, sub_prop;
    char *query;
    char *sconstr = NULL;

    assert(rule == 2 || rule == 3);

    /* Only delete if were doing all models, as the a-box model has allready
     * been flushed, or the KB is monotonically increasing, so there wont be
     * any removals */
    if (!by_model) {
	query = g_strdup_printf("DELETE FROM "TBLPREFIX"triples WHERE inferred=%d", rule);
	if (mysql_query(db->db, query)) {
	    fprintf(stderr, "Failed to remove previous entailments ('%s', %d): %s\n", query, mysql_errno(db->db), mysql_error(db->db));
	    return;
	}
	free(query);
    }
	
    switch (rule) {
	case 2:
	    predicate = rs_hash(RDFS_DOMAIN);
	    break;
	case 3:
	    predicate = rs_hash(RDFS_RANGE);
	    break;
	default:
	    fprintf(stderr, "This should never happen %s:%d\n", __FILE__,
		    __LINE__);
	    return;
    }
    type = rs_hash(RDF_TYPE);
    sub_prop = rs_hash(RDFS_SUBPROPERTY);
   
    mysql_query(db->db, "DROP TABLE tmp_mso");
    if (mysql_query(db->db, TMP_MSO_TABLE)) {
	fprintf(stderr, "Failed to create MSO table (%d): %s\n", mysql_errno(db->db), mysql_error(db->db));
	return;
    }

    if (by_model) {
	sconstr = g_strdup_printf(" AND t1.model=%lld", smodel);
    }
    fflush(stdout);
    if (rule == 2) {
	query = g_strdup_printf("INSERT INTO tmp_mso SELECT t1.model, t1.subject, t3.object FROM "TBLPREFIX"triples t1, "TBLPREFIX"triples t2, "TBLPREFIX"triples t3 WHERE t1.predicate=t2.subject AND t2.predicate=%lld AND t2.object=t3.subject AND t3.predicate=%lld%s", sub_prop, predicate, by_model ? sconstr : ""); 
    } else {
	query = g_strdup_printf("INSERT INTO tmp_mso SELECT t1.model, t1.object, t3.object FROM "TBLPREFIX"triples t1, "TBLPREFIX"triples t2, "TBLPREFIX"triples t3 WHERE t1.predicate=t2.subject AND t2.predicate=%lld AND t2.object=t3.subject AND t3.predicate=%lld%s", sub_prop, predicate, by_model ? sconstr : ""); 
    }
    if (mysql_query(db->db, query)) {
	if (mysql_errno(db->db) != 1062) {
	    fprintf(stderr, "Failed to execute query '%s' (%d): %s\n", query, mysql_errno(db->db), mysql_error(db->db));
	    return;
	}
    }
    free(query);
    free(sconstr);

    query = g_strdup_printf("INSERT INTO "TBLPREFIX"triples SELECT m, s, %lld, o, 0, %d FROM tmp_mso", type, rule);
    if (mysql_query(db->db, query)) {
	fprintf(stderr, "Failed to transfer results '%s' (%d): %s\n", query, mysql_errno(db->db), mysql_error(db->db));
    }
    free(query);

    mysql_query(db->db, "DROP TABLE tmp_mso");
}

/* rs_rdfs4: cacluate and assert RDFS4a+b entailments */
static void rs_rdfs4(RDFSQL *db)
{
    guint64 model, type, resource;
    char *query;

    model = rs_check_model(db, "3store:cache#RDFS4");
    rs_flush_other(db, model);

    type = rs_hash(RDF_TYPE);
    resource = rs_hash(RDFS_RESOURCE);

    query = g_strdup_printf("INSERT INTO "TBLPREFIX"triples SELECT %lld, hash, %lld, %lld, 0, 1 FROM "TBLPREFIX"resources", model, type, resource);
    if (mysql_query(db->db, query)) {
	fprintf(stderr, "Failed to get RDFS4 entailments with '%s' (%d): %s\n", query, mysql_errno(db->db), mysql_error(db->db));
    }
    free(query);
}

/* rs_entail_rdfs2: generate entailments for a partiular model, designed to be
 * called externally. */
void rs_entail_rdfs2(RDFSQL *db, guint64 model)
{  
    rs_rdfs23(db, 2, 1, model);
}

/* rs_entail_rdfs3: generate entailments for a partiular model, designed to be
 * called externally. */
void rs_entail_rdfs3(RDFSQL *db, guint64 model)
{  
    rs_rdfs23(db, 3, 1, model);
}

/* vi:set ts=8 sts=4 sw=4: */
