/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: import.c,v 1.8 2005/02/16 12:12:30 theno23 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <mysql.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdarg.h> /* XXX raptor 1.3.0 needs this - can be removed later */
#include <raptor.h>
#include <getopt.h>
#include <glib.h>

#include "config.h"
#include "rdfsql/rdfsql.h"

#define INT_NS "http://triplestore.aktors.org/ontology/#"

#define MEMBER_PREFIX "http://www.w3.org/1999/02/22-rdf-syntax-ns#_"
#define MEMBER_PROP   "http://www.w3.org/2000/01/rdf-schema#member"

#define INTEGER_DT    "http://www.w3.org/2001/XMLSchema#integer"

void store_stmt(void *user_data, const raptor_statement * statement);

static void fatal_error_handler(void *data, raptor_locator * locator,
                          const char *message);
static void error_handler(void *data, raptor_locator * locator,
                          const char *message);

static void warning_handler(void *data, raptor_locator * locator,
                            const char *message);

int rs_import2(RDFSQL *db, char *uri, char *model, int verbosity, FILE *msg,
               rs_syntax format);

static int raptor_inited = 0;

static unsigned int uid = 0;    /* A unique(ish) id to append to genid's to
                                 * help avoid clashses */

typedef struct {
    int verbosity;
    int count_err;
    int count_warn;
    RDFSQL *db;
    FILE *msg;
} rs_parse_stuff;

int rs_import(RDFSQL *db, char *uri, char *model, int verbosity)
{
    return rs_import2(db, uri, model, verbosity, stderr, RS_AUTO);
}

int rs_import2(RDFSQL *db, char *uri, char *model, int verbosity, FILE *msg,
               rs_syntax format)
{
    char timestamp[256];
    char new_model_uri[256];
    time_t t;
    struct tm *tp;
    guint64 old_model;
    char *poll_obj_uri;
    char *query;
    raptor_parser *rdf_parser = NULL;
    raptor_uri ruri, muri;
    rs_parse_stuff parse_data;

    parse_data.verbosity = verbosity;
    parse_data.db = db;
    parse_data.msg = msg;

    old_model = db->model;

    if (!uid) {
	struct timeval tv;

	gettimeofday(&tv, NULL);
	// XXX should be replaced with a table entry
	uid = (unsigned int) getpid();
	uid ^= (unsigned int) tv.tv_usec;
    } else {
	uid++;
    }
    poll_obj_uri = g_strdup_printf("3store:po:%08x", uid);

    t = time(NULL);
    tp = localtime(&t);
    snprintf(timestamp, 255, "%04d-%02d-%02dT%02d:%02d:%02d", tp->tm_year +
	     1900, tp->tm_mon + 1, tp->tm_mday, tp->tm_hour, tp->tm_min,
	     tp->tm_sec);

    if (!raptor_inited) {
        raptor_init();
	raptor_inited = 1;
    }

    if (format == RS_AUTO) {
	if (strstr(uri, ".n3") || strstr(uri, ".ttl")) {
	    if (verbosity > 0) fprintf(msg, "using Turtle parser\n");
	    rdf_parser = raptor_new_parser("turtle");
	} else if (strstr(uri, ".nt")) {
	    if (verbosity > 0) fprintf(msg, "using NTriples parser\n");
	    rdf_parser = raptor_new_parser("ntriples");
	} else {
	    if (verbosity > 0) fprintf(msg, "using RDF/XML parser\n");
	    rdf_parser = raptor_new_parser("rdfxml");
	}
    } else if (format == RS_XML) {
	if (verbosity > 0) fprintf(msg, "using RDF/XML parser\n");
	rdf_parser = raptor_new_parser("rdfxml");
    } else if (format == RS_NTRIPLES) {
	if (verbosity > 0) fprintf(msg, "using NTriples parser\n");
	rdf_parser = raptor_new_parser("ntriples");
    } else if (format == RS_TURTLE) {
	if (verbosity > 0) fprintf(msg, "using Turtle parser\n");
	rdf_parser = raptor_new_parser("turtle");
    }
    if (!rdf_parser) {
	fprintf(msg, "Failed to create parser\n");
	return 1;
    }

    raptor_set_statement_handler(rdf_parser, &parse_data, store_stmt);
    raptor_set_warning_handler(rdf_parser, &parse_data, warning_handler);
    raptor_set_error_handler(rdf_parser, &parse_data, error_handler);
    raptor_set_fatal_error_handler(rdf_parser, &parse_data, fatal_error_handler);

    ruri = raptor_new_uri(uri);
    muri = raptor_new_uri(model);

    db->model = rs_check_model(db, model);
    parse_data.count_err = 0;
    parse_data.count_warn = 0;
    if (raptor_parse_uri(rdf_parser, ruri, muri)) {
	fprintf(msg, "Failed to parse file\n");
	if (mysql_query(db->db, "UNLOCK TABLES")) {
	    fprintf(msg, "Failed to unlock tables (%d): %s\n",
		    mysql_errno(db->db), mysql_error(db->db));
	}
	return 1;
    }

    snprintf(new_model_uri, 255, "3store:metadata:%llx", db->model);
    db->model = rs_check_model(db, new_model_uri);

    query = g_strdup_printf("DELETE FROM "TBLPREFIX"triples WHERE subject=%lld AND "
		    "predicate=%lld", old_model,
		    rs_hash(INT_NS "last-polled"));
    if (mysql_query(db->db, query)) {
	fprintf(msg, "Failed to remove old timestamp '%s' (%d): %s\n",
		query, mysql_errno(db->db), mysql_error(db->db));
    }
    free(query);

    rs_assert_triple(db, model, "http://purl.org/dc/elements/1.1/source", uri,
		     ObjURI);
    rs_assert_triple(db, model, INT_NS "last-polled", poll_obj_uri, ObjURI);
    rs_assert_triple(db, model, INT_NS "polled", poll_obj_uri, ObjURI);
    rs_assert_triple(db, poll_obj_uri, INT_NS "date", timestamp,
		     ObjLiteral);
    if (parse_data.count_err == 0 && parse_data.count_warn == 0) {
	rs_assert_triple(db, poll_obj_uri, INT_NS "status-message", "OK",
			 ObjLiteral);
	rs_assert_triple(db, poll_obj_uri, INT_NS "status",
			 INT_NS "success", ObjURI);
    } else {
	char msg[256];
	snprintf(msg, 255, "%d error%s and %d warning%s",
		 parse_data.count_err, parse_data.count_err == 1 ? "" : "s",
		 parse_data.count_warn, parse_data.count_warn == 1 ?  "" : "s");
	rs_assert_triple(db, poll_obj_uri, INT_NS "status-message", msg,
			 ObjLiteral);
	if (parse_data.count_err > 0) {
	    rs_assert_triple(db, poll_obj_uri, INT_NS "status",
			     INT_NS "failure", ObjURI);
	} else {
	    rs_assert_triple(db, poll_obj_uri, INT_NS "status",
			     INT_NS "warning", ObjURI);
	}
    }
    db->model = old_model;

    return parse_data.count_err;
}

void store_stmt(void *user_data, const raptor_statement * statement)
{
    rs_parse_stuff *data = (rs_parse_stuff *)user_data;
    rs_obj_type type;
    char *tmps, *tmpp, *tmpo;
    char *subj = raptor_uri_as_string((raptor_uri *)statement->subject),
	 *pred = raptor_uri_as_string((raptor_uri *)statement->predicate),
         *obj  = raptor_uri_as_string((raptor_uri *)statement->object);
    const char *datatype =
	raptor_uri_as_string(statement->object_literal_datatype);

    if (statement->object_type == RAPTOR_IDENTIFIER_TYPE_LITERAL ||
	statement->object_type == RAPTOR_IDENTIFIER_TYPE_XML_LITERAL) {
	type = ObjLiteral;
    } else {
	type = ObjURI;
    }
    if (statement->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	tmps = g_strdup_printf("_:%s.%x", (char *) statement->subject, uid);
	subj = tmps;
    } else {
	tmps = NULL;
    }
    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	tmpp = g_strdup_printf("_:%s.%x", (char *) statement->predicate, uid);
	pred = tmpp;
    } else {
	tmpp = NULL;
    }
    if (statement->object_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	tmpo = g_strdup_printf("_:%s.%x", (char *) statement->object, uid);
	obj = tmpo;
    } else {
	tmpo = NULL;
    }

    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
	tmpp = g_strdup_printf(MEMBER_PREFIX "%d", *((int *)pred));
	pred = tmpp;
    }

    if (data->verbosity > 1) {
	fprintf(stdout, "asserting (<%s>, <%s>, %c%s%c)\n", subj, pred,
	       type == ObjLiteral ? '"' : '<', obj,
	       type == ObjLiteral ? '"' : '>');
    }

    if (rs_assert_triple(data->db, subj, pred, obj, type) == RS_BAD_SQL) {
	fprintf(data->msg, "tstore_import: Bad SQL, skipping statement\n");
    }

    /* old string prefix check */
    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL || !strncmp(pred, MEMBER_PREFIX, 44)) {
	if (rs_assert_triple(data->db, subj, MEMBER_PROP, obj, type) ==
	    RS_BAD_SQL) {
	    fprintf(data->msg, "tstore_import: Bad SQL, skipping entailment\n");
	}
	if (rs_assert_triple(data->db, pred, RDFS_SUBPROPERTY, MEMBER_PROP,
			     ObjURI) == RS_BAD_SQL) {
	    fprintf(data->msg, "tstore_import: Bad SQL, skipping entailment\n");
	}
    }

    if (datatype && !strcmp(datatype, INTEGER_DT)) {
	char *canonical = g_strdup_printf("%d", atoi(obj));

	if (strcmp(obj, canonical) &&
            rs_assert_triple(data->db, subj, pred, canonical, type) ==
	    RS_BAD_SQL) {
	    fprintf(data->msg, "tstore_import: Bad SQL, skipping entailment\n");
	}
	free(canonical);
    }

    free(tmps);
    free(tmpp);
    free(tmpo);
}

static void
fatal_error_handler(void *user_data, raptor_locator * locator,
                    const char *message)
{
    rs_parse_stuff *data = (rs_parse_stuff *)user_data;

    fprintf(data->msg, "3store import fatal error: ");
    raptor_print_locator(stderr, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_err)++;
}

static void
error_handler(void *user_data, raptor_locator * locator, const char *message)
{
    rs_parse_stuff *data = (rs_parse_stuff *)user_data;

    fprintf(data->msg, "3store import error: ");
    raptor_print_locator(stderr, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_err)++;
}

static void
warning_handler(void *user_data, raptor_locator * locator, const char *message)
{
    rs_parse_stuff *data = (rs_parse_stuff *)user_data;

    fprintf(data->msg, "3store import warning: ");
    raptor_print_locator(stderr, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_warn)++;
}

void rs_update_caches(RDFSQL *db, int verbosity)
{
    MYSQL_RES *res;
    MYSQL_ROW row;

    if (mysql_query(db->db, "SELECT taxonomy FROM "TBLPREFIX"cache_state LIMIT 1")) {
        fprintf(stderr, "Failed to read cache state (%d): %s\n",
                mysql_errno(db->db), mysql_error(db->db));
    }

    res = mysql_store_result(db->db);
    row = mysql_fetch_row(res);
    mysql_free_result(res);

    if (row && atoi(row[0])) {
        if (verbosity > 0) {
            printf("Taxonomy has changed, rebuilding lattices\n");
        }
        /* Something to do with the schema was changed, rebuild all the
         * entailments */
        rs_rebuild_taxonomy(db, verbosity);
    } else {
        /* Nothing in the schema has changed, but we need to do 2+3 anyway */
        if (verbosity > 0) {
            printf("Finding rdfs2 entailments\n");
        }
        rs_entail_rdfs2(db, db->model);
        if (verbosity > 0) {
            printf("Finding rdfs3 entailments\n");
        }
        rs_entail_rdfs3(db, db->model);
    }
}

/* vi:set ts=8 sts=4 sw=4: */
