/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: rdfsql.c,v 1.25 2005/03/22 19:21:17 theno23 Exp $
 */

#include <glib.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <mysql.h>
#include <mysqld_error.h>

#include "config.h"
#include "rdfsql/rdfsql.h"
#include "md5.h"

#define mysql_equery(d, q) if (mysql_query(d, q)) { fprintf(stderr, "SQL error in '%s' %d: %s\n", q, mysql_errno(this->db), mysql_error(this->db)); return RS_BAD_SQL; }
#define check_buffer(s) if (strlen(s)*2 + 1 > buffer_len) {buffer_len = strlen(s)*2 + 100; buffer = realloc(buffer, buffer_len);}

guint64 rs_ensure_resource(const RDFSQL *this, const char *uri);

/* rs_hash:
 *
 * A hasing function to return a 64bit hash from a NULL terminated string.
 * Internally the function is the top 64 bits of a 128bit MD5 checksum.
 */

guint64 rs_hash(const char *str)
{
    guint64 data[2];

    if (!str) {
	return 0;
    }
    md5_buffer(str, strlen(str), data);

    return GUINT64_FROM_BE(data[0]);
}

RDFSQL *rs_connect_remote(const char *host, const char *db,
			  const char *user, const char *passwd,
			  const char *model)
{
    RDFSQL *new = calloc(1, sizeof(RDFSQL));

    if (!(new->db = mysql_init(NULL))) {
	fprintf(stderr, "Could not initialise database connection\n");
	return NULL;
    }
    mysql_options(new->db, MYSQL_READ_DEFAULT_GROUP, "3store");
    if (!mysql_real_connect
	(new->db, host, user, passwd, db, 0, NULL, 0)) {
	fprintf(stderr, "Failed to connect to database: Error: %s\n",
		mysql_error(new->db));
	return NULL;
    }

    if (model) {
	new->model = rs_check_model(new, model);
    }

    return new;
}

RDFSQL *rs_connect(const char *db, const char *user, const char *passwd,
		   const char *model)
{
    return rs_connect_remote("localhost", db, user, passwd, model);
}

void rs_disconnect(RDFSQL *rs)
{
    mysql_close((MYSQL *)rs->db);
}

guint64 rs_check_model(RDFSQL * this, const char *model)
{
    guint64 hash = rs_hash(model);
    char tmp[512], buffer[512];

    mysql_real_escape_string(this->db, buffer, model, strlen(model));
    snprintf(tmp, 511, "INSERT INTO %smodels VALUES(%lld, '%s')", TBLPREFIX, hash, buffer);
    if (mysql_query(this->db, tmp)) {
	if (mysql_errno(this->db) != ER_DUP_ENTRY) {
	    fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		    mysql_errno(this->db), mysql_error(this->db));
	    return 0;
	}
    }

    return hash;
}

void rs_flush(RDFSQL * this)
{
    rs_flush_other(this, this->model);
}

void rs_flush_other(RDFSQL * this, guint64 model)
{
    char *tmp = g_strdup_printf("DELETE FROM " TBLPREFIX "triples WHERE model=%lld",
				model);

    if (mysql_query(this->db, tmp)) {
	fprintf(stderr, "SQL error in flush: '%s' %d: %s\n", tmp,
		mysql_errno(this->db), mysql_error(this->db));
    }
}

void rs_optimize_tables(RDFSQL * this, int verbosity)
{
    const char *tables[] = {
	TBLPREFIX "triples",
	TBLPREFIX "resources",
	TBLPREFIX "literals",
	TBLPREFIX "models",
	NULL
    };
    unsigned int i;
    MYSQL_ROW row;
    MYSQL_RES *res;
    char tmp[256];

    for (i = 0; tables[i]; i++) {
	  snprintf(tmp, 255, "OPTIMIZE TABLE %s",tables[i]);
	if (mysql_query(this->db, tmp)) {
	    fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		    mysql_errno(this->db), mysql_error(this->db));
	} else {
	    res = mysql_store_result(this->db);
	    while ((row = mysql_fetch_row(res))) {
		if (verbosity > 0) {
		    printf("%s: %s\n", row[0], row[3]);
		}
	    }
	    mysql_free_result(res);
	}
    }
}

int rs_assert_triple_intl(const RDFSQL *this, const char *subj,
			  const char *pred, const char *obj,
			  const rs_obj_type type, int inferred);

int rs_assert_triple_intl(const RDFSQL *this, const char *subj,
			  const char *pred, const char *obj,
			  const rs_obj_type type, int inferred)
{
    char tmp[512];
    MYSQL_RES *res;
    MYSQL_ROW row;
    static char *buffer = NULL;
    static char *lbuffer = NULL;
    static unsigned int lbuffer_len = 0;
    guint64 s, p, o;

    if (type == ObjAny) {
	return RS_BAD_TYPE;
    }

    s = rs_ensure_resource(this, subj);

    p = rs_ensure_resource(this, pred);
    if (type == ObjURI) {
	o = rs_ensure_resource(this, obj);
    } else {
	o = rs_hash(obj);
	snprintf(tmp, 256, "SELECT count(*) FROM "TBLPREFIX"literals WHERE hash=%lld",
		 o);
	mysql_equery(this->db, tmp);
	res = mysql_store_result(this->db);
	row = mysql_fetch_row(res);
	mysql_free_result(res);

	/* if the count(*) returned 0, insert the literal */
	if (*(char *) row[0] == '0') {
	    buffer = malloc(strlen(obj) * 2 + 1);
	    mysql_real_escape_string(this->db, buffer, obj, strlen(obj));
	    if (lbuffer_len < strlen(buffer) + 100) {
		lbuffer_len = strlen(buffer) + 100;
		lbuffer = realloc(lbuffer, lbuffer_len);
	    }
	    /* Must not be delayed as we check this one by hand */
	    sprintf(lbuffer, "INSERT INTO "TBLPREFIX"literals VALUES(%lld, '%s')", o,
		    buffer);
	    if (mysql_query(this->db, lbuffer)) {
		if (mysql_errno(this->db) != ER_DUP_ENTRY) {
		    fprintf(stderr, "SQL error %d in '%s': %s\n",
			    mysql_errno(this->db), tmp,
			    mysql_error(this->db));
		    return RS_BAD_SQL;
		}
	    }
	}
    }

    return rs_assert_raw(this, s, p, o, type, inferred);
}

gboolean cull_old(gpointer key, gpointer value, gpointer user_data)
{
    if (*((int *)value) < (int)user_data) {
	free(key);
	free(value);
	return 1;
    }

    return 0;
}

guint64 rs_ensure_resource(const RDFSQL *this, const char *uri)
{
    guint64 hash;
    static GHashTable *recent = NULL;
    static unsigned int recent_tick = 0;
    static unsigned int last_cleanup = 0;
    unsigned int *when;
    char *buffer;
    char tmp[512];

    recent_tick++;
    if (recent_tick > last_cleanup + 1024) {
	g_hash_table_foreach_remove(recent, cull_old,
				    (gpointer)(recent_tick - 768));
	last_cleanup = recent_tick;
    }

    hash = rs_hash(uri);
    if (!recent) {
	recent = g_hash_table_new(g_str_hash, g_str_equal);
    }
    when = (unsigned int *)g_hash_table_lookup(recent, uri);
    if (when) {
	*when = recent_tick;

	return hash;
    }

    when = malloc(sizeof(int));
    *when = recent_tick;
    g_hash_table_insert(recent, strdup(uri), when);

    buffer = malloc(strlen(uri) * 2 + 1);
    mysql_real_escape_string(this->db, buffer, uri, strlen(uri));
    snprintf(tmp, 511, "INSERT INTO "TBLPREFIX"resources VALUES(%lld, '%s')", hash,
	     buffer);
    if (mysql_query(this->db, tmp)) {
	if (mysql_errno(this->db) != ER_DUP_ENTRY) {
	    fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		    mysql_errno(this->db), mysql_error(this->db));
	    return RS_BAD_SQL;
	}
    }

    return hash;
}

int rs_assert_raw(const RDFSQL * this, guint64 subj, guint64 pred,
		  guint64 obj, const rs_obj_type type, int inferred)
{
    char tmp[256];

    snprintf(tmp, 256,
	     "INSERT INTO "TBLPREFIX"triples VALUES(%lld, %lld, %lld, %lld, %d, %d)",
	     this->model, subj, pred, obj, type, inferred);
    if (mysql_query(this->db, tmp)) {
	if (mysql_errno(this->db) == ER_DUP_ENTRY) {
	    /* triple is allready asserted, quit */
	    return RS_ALLREADY;
	} else {
	    /* other SQL error */
	    fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		    mysql_errno(this->db), mysql_error(this->db));
	    return RS_BAD_SQL;
	}
    }

    return 0;
}

int rs_assert_raw_block(const RDFSQL * this, rs_raw_triple * block,
			int count, const rs_obj_type type, int inferred)
{
    char *query;
    int ofs;
    int length;
    int i;

    length = 30 + 110 * count;
    query = malloc(length * sizeof(char));
    ofs = sprintf(query, "INSERT INTO "TBLPREFIX"triples VALUES ");
    for (i = 0; i < count; i++) {
	ofs +=
	    snprintf(query + ofs, 110, "(%lld,%lld,%lld,%lld,%d,%d)%s",
		     this->model, block[i].subj, block[i].pred,
		     block[i].obj, type, inferred,
		     i == count - 1 ? "" : ",");
    }
    if (ofs >= length) {
	fprintf(stderr, "Buffer overrun in %s:%d\n", __FILE__, __LINE__);
	return RS_BAD_SQL;
    }
    if (mysql_query(this->db, query)) {
	if (mysql_errno(this->db) == ER_DUP_ENTRY) {
	    /* one triple is allready asserted, we're going to have
	     * to do this by hand :( */
	    for (i = 0; i < count; i++) {
		rs_assert_raw(this, block[i].subj,
			      block[i].pred, block[i].obj, type, inferred);
	    }
	} else {
	    /* other SQL error */
	    fprintf(stderr, "SQL error in '%s' %d: %s\n", query,
		    mysql_errno(this->db), mysql_error(this->db));
	    return RS_BAD_SQL;
	}
    }
    free(query);

    return 0;
}

int rs_assert_triple(const RDFSQL * this, const char *subj,
		     const char *pred, const char *obj,
		     const rs_obj_type type)
{
    int ret = rs_assert_triple_intl(this, subj, pred, obj, type, 0);

    if (!strcmp(pred, RDFS_SUBCLASS) || !strcmp(pred, RDFS_SUBPROPERTY)) {
	if (mysql_query
	    (this->db, "UPDATE "TBLPREFIX"cache_state SET taxonomy=taxonomy+1")) {
	    fprintf(stderr, "SQL error in cache state update %d: %s\n",
		    mysql_errno(this->db), mysql_error(this->db));
	    return RS_BAD_SQL;
	}
    }

    return ret;
}

int rs_assert_inferred_triple(const RDFSQL * this, const char *subj,
			      const char *pred, const char *obj,
			      const rs_obj_type type)
{
    return rs_assert_triple_intl(this, subj, pred, obj, type, 1);
}

rs_result *rs_find_triples_intl(RDFSQL * this, const char *subj,
				const char *pred, const char *obj,
				const rs_obj_type type,
				int inferred,
				int num_triples, const char *model);


rs_result *rs_find_triples_intl(RDFSQL * this, const char *subj,
				const char *pred, const char *obj,
				const rs_obj_type type,
				int inferred,
				int num_triples, const char *model)
{
    char tmp[1024] = "SELECT ";
    char hash[64];
    int need_joiner = 0;
    int idx = 0;
    rs_result *res;

    res = calloc(1, sizeof(rs_result));
    if (!subj) {
	strcat(tmp, "rs.uri as subject");
	need_joiner = 1;
    }
    if (!pred) {
	if (need_joiner) {
	    strcat(tmp, ", ");
	}
	strcat(tmp, "rp.uri as predicate");
	need_joiner = 1;
    }
    if (!obj) {
	if (need_joiner) {
	    strcat(tmp, ", ");
	}
	if (type == ObjAny) {
	    strcat(tmp, "ro.uri as object_uri, lo.literal as object_lit");
	} else if (type == ObjURI) {
	    strcat(tmp, "ro.uri as object");
	} else {
	    strcat(tmp, "lo.literal as object");
	}
    }
    if (type == ObjAny) {
	strcat(tmp, ", "TBLPREFIX"triples.literal");
    }

    strcat(tmp, " FROM "TBLPREFIX"triples");
    if (!subj) {
	strcat(tmp, ", "TBLPREFIX"resources rs");
    }
    if (!pred) {
	strcat(tmp, ", "TBLPREFIX"resources rp");
    }
    if (!obj) {
	if (type == ObjAny) {
	    strcat(tmp,
		   " LEFT JOIN "TBLPREFIX"literals lo ON "TBLPREFIX"triples.object=lo.hash");
	    strcat(tmp,
		   " LEFT JOIN "TBLPREFIX"resources ro ON "TBLPREFIX"triples.object=ro.hash");
	    strcat(tmp, " WHERE ");
	} else if (type == ObjLiteral) {
	    strcat(tmp, ", "TBLPREFIX"literals lo WHERE "TBLPREFIX"triples.object=lo.hash AND ");
	} else if (type == ObjURI) {
	    strcat(tmp,
		   ", "TBLPREFIX"resources ro WHERE "TBLPREFIX"triples.object=ro.hash AND ");
	}
    } else {
	strcat(tmp, " WHERE ");
    }
    need_joiner = 0;
    if (!subj) {
	strcat(tmp, TBLPREFIX"triples.subject=rs.hash");
	need_joiner = 1;
    }
    if (!pred) {
	if (need_joiner) {
	    strcat(tmp, " AND ");
	}
	strcat(tmp, TBLPREFIX"triples.predicate=rp.hash");
	need_joiner = 1;
    }
    if (subj) {
	if (need_joiner) {
	    strcat(tmp, " AND ");
	}
	sprintf(hash, "%lld", rs_hash(subj));
	strcat(tmp, TBLPREFIX"triples.subject=");
	strcat(tmp, hash);
	res->subj_idx = -1;
	res->t.subject = strdup(subj);
	need_joiner = 1;
    } else {
	res->subj_idx = idx++;
    }
    if (pred) {
	if (need_joiner) {
	    strcat(tmp, " AND ");
	}
	sprintf(hash, "%lld", rs_hash(pred));
	strcat(tmp, TBLPREFIX"triples.predicate=");
	strcat(tmp, hash);
	res->pred_idx = -1;
	res->t.predicate = strdup(pred);
	need_joiner = 1;
    } else {
	res->pred_idx = idx++;
    }
    if (obj) {
	sprintf(hash, "%lld", rs_hash(obj));
	strcat(tmp, " AND "TBLPREFIX"triples.object=");
	strcat(tmp, hash);
	res->obj_idx = -1;
	res->t.object = strdup(obj);
    } else {
	res->obj_idx = idx++;
    }
    if (type == ObjAny) {
	res->lit_idx = idx + 1;
    } else {
	strcat(tmp, " AND "TBLPREFIX"triples.literal=");
	sprintf(hash, "%d", (int) type);
	strcat(tmp, hash);
	res->lit_idx = -1;
	if (type == ObjLiteral) {
	    res->t.literal = 1;
	}
    }
    if (model) {
	strcat(tmp, " AND "TBLPREFIX"triples.model=");
	sprintf(hash, "%lld", rs_hash(model));
	strcat(tmp, hash);
    }

    if (inferred == 0) {
	strcat(tmp, " AND "TBLPREFIX"triples.inferred=0");
    }
    if (num_triples) {
	strcat(tmp, " LIMIT ");
	sprintf(hash, "%d", num_triples);
	strcat(tmp, hash);
    }
// printf("Q: %s\n", tmp);

    if (mysql_query(this->db, tmp)) {
	fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		mysql_errno(this->db), mysql_error(this->db));
	return NULL;
    }
    res->db_res = mysql_store_result(this->db);

    return res;
}

rs_result *rs_find_triples(RDFSQL * this, const char *subj,
			   const char *pred, const char *obj,
			   const rs_obj_type type, int num_triples,
			   const char *model)
{
    return rs_find_triples_intl(this, subj, pred, obj, type, 1,
				num_triples, model);
}

rs_result *rs_find_asserted_triples(RDFSQL * this, const char *subj,
				    const char *pred, const char *obj,
				    const rs_obj_type type,
				    int num_triples, const char *model)
{
    return rs_find_triples_intl(this, subj, pred, obj, type, 0,
				num_triples, model);
}

int rs_num_triples(rs_result * res)
{
    return mysql_num_rows((MYSQL_RES *) (res->db_res));
}

rs_triple *rs_next_triple(rs_result * res)
{
    MYSQL_ROW row;

    if (!res || !res->db_res) {
	return NULL;
    }

    row = mysql_fetch_row((MYSQL_RES *) (res->db_res));

    if (!row) {
	return NULL;
    }

    if (res->subj_idx >= 0) {
	res->t.subject = row[res->subj_idx];
    }
    if (res->pred_idx >= 0) {
	res->t.predicate = row[res->pred_idx];
    }
    if (res->obj_idx >= 0) {
	if (res->lit_idx >= 0 && *(char *) row[res->lit_idx] == '1') {
	    res->t.object = row[res->obj_idx + 1];
	    res->t.literal = 1;
	} else {
	    res->t.object = row[res->obj_idx];
	    res->t.literal = 0;
	}
    }

    return &(res->t);
}

void rs_free_result(rs_result * res)
{
    mysql_free_result((MYSQL_RES *) res->db_res);
    if (res->subj_idx == -1) {
	free(res->t.subject);
    }
    if (res->pred_idx == -1) {
	free(res->t.predicate);
    }
    if (res->obj_idx == -1) {
	free(res->t.object);
    }
    free(res);
}

rs_result *rs_find_all_resources(const RDFSQL * this, int limit,
				 const char *model)
{
    rs_result *res;
    char tmp[256];
    char tmp2[64];

    sprintf(tmp,
	    "SELECT uri FROM "TBLPREFIX"resources,"TBLPREFIX"triples WHERE "TBLPREFIX"resources.hash=triples.subject");
    if (model) {
	sprintf(tmp2, " AND model=%lld", rs_hash(model));
	strcat(tmp, tmp2);
    }
    if (limit) {
	sprintf(tmp2, " LIMIT %d", limit);
	strcat(tmp, tmp2);
    }

    if (mysql_query(this->db, tmp)) {
	fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		mysql_errno(this->db), mysql_error(this->db));
	return NULL;
    }

    res = calloc(1, sizeof(rs_result));
    res->db_res = mysql_store_result(this->db);
    res->subj_idx = 0;
    res->t.subject = NULL;
    res->pred_idx = -2;
    res->t.predicate = NULL;
    res->obj_idx = -2;
    res->t.object = NULL;

    return res;
}

rs_result *rs_find_all_models(const RDFSQL * this)
{
    rs_result *res;
    char tmp[256];

    sprintf(tmp, "SELECT model FROM %smodels",TBLPREFIX);

    if (mysql_query(this->db, tmp)) {
	fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		mysql_errno(this->db), mysql_error(this->db));
	return NULL;
    }

    res = calloc(1, sizeof(rs_result));
    res->db_res = mysql_store_result(this->db);
    res->subj_idx = 0;
    res->t.subject = NULL;
    res->pred_idx = -2;
    res->t.predicate = NULL;
    res->obj_idx = -2;
    res->t.object = NULL;

    return res;
}

/* vi:set ts=8 sts=4 sw=4: */
