#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <pthread.h>
#include <rdfsql/rdfsql.h>
#include <mysql/mysql.h>

#include "config.h"
#include "lex.h"
#include "rdql_parser.h"
#include "rdfsql/rdql.h"
#include "rdfsql/rdql_datatypes.h"

pthread_mutex_t yacc_m = PTHREAD_MUTEX_INITIALIZER;

char *y_err_str = NULL;

void var_walk(gpointer key, gpointer value, gpointer user_data);
char *strdup_escape(MYSQL *mysql, const char *s);

static void st_free(gpointer key, gpointer value, gpointer ud)
{
    rdql_variable *v = value;
    if (v->column) free(v->column);
    if (v->val_column) free(v->val_column);
    if (v->hash_column) free(v->hash_column);
    free(v);
}

rs_rdql_result *rs_rdql_query(RDFSQL * this, const char *query_str, int
		max_comp)
{
    void *bs;
    query_t query;
    GSList *li;
    unsigned int i = 0;
    int uid = 1024;
    GHashTable *sym_tbl = g_hash_table_new(g_str_hash, g_str_equal);
    int ret_vars = 0;
    int triples = 0;
    int constraints = 0;
    int col_cnt = 0;
    int tmp_cnt = 0;
    int complexity = 0;
    strvec_t *selects, *froms, *wheres, *joins;
    char *select = NULL, *from = NULL, *where = NULL, *join = NULL;
    char *sql_query;
    char *sql_explain;
    char *limit = NULL;
    char tmp[512];
    rs_rdql_result *res;
    rs_rdql_column *cols;
    MYSQL_ROW row;

    pthread_mutex_lock(&yacc_m);
    /* XXX start of un-threadsafe code */
    y_err_str = NULL;
    bs = yy_scan_string(query_str);
    yy_switch_to_buffer(bs);
    yyparse();
    memcpy(&query, &q_data, sizeof(query));
    /* XXX end of un-threadsafe code */
    pthread_mutex_unlock(&yacc_m);
    if (y_err_str) {
	g_hash_table_destroy(sym_tbl);

	return rdql_error(y_err_str);
    }

    /* Add standard prefixes */
    query.prefixes = g_slist_append(query.prefixes,
				    new_prefix("rdf", RDF_NAMESPACE));
    query.prefixes = g_slist_append(query.prefixes,
				    new_prefix("rdfs", RDFS_NAMESPACE));
    query.prefixes = g_slist_append(query.prefixes,
				    new_prefix("owl", OWL_NAMESPACE));

    if (query.limit) {
	limit = g_strdup_printf(" LIMIT %d, %d", query.offset, query.limit);
    } else if (query.offset) {
	limit = g_strdup_printf(" LIMIT %d, -1", query.offset);
    }

    for (li = query.select, i = 0; li; li = li->next, i++) {
	rdql_variable *v;

	v = constrain_var_type(sym_tbl, (char *) li->data, ObjAny);
	v->returned = i + 1;
	ret_vars++;
    }

    for (li = query.where, i = 0; li; li = li->next, i++) {
	pattern_t *p = li->data;
	rdql_variable *v;

	expand_pattern(p, query.prefixes);
	triples++;
	constraints += 3;
	if (p->m) constraints++;

	if (p->p->type == URI && !strcmp(p->p->value, RDF_TYPE)) {
	    /* Rewrite the triple to be indirected and append new
	     * triple to end of list eg.
	     *
	     * (?a, <rdf:type>, <foo:bar>) =>
	     *   (?a, <rdf:type>, ?foo),
	     *   (?foo, <rdf:subClassOf>, <foo:bar>)
	     */
	    part_t *new_s, *new_p, *new_o;
	    pattern_t *new_pa;

	    new_s = new_part(VARIABLE, g_strdup_printf("_tmp%d", tmp_cnt));
	    new_p = new_part(URI, strdup(RDFS_SUBCLASS));
	    new_o = new_part(VARIABLE, g_strdup_printf("_tmp%d",
						       tmp_cnt++));
	    new_pa = new_pattern(new_s, new_p, p->o, NULL);
	    p->o = new_o;
	    g_slist_append(query.where, new_pa);
	} else if (p->p->type == URI) {
	    /* Rewrite the triple to use subproperty inference if
	     * we think there are subproperties:
	     *
	     * (?a, <foo:prop>, ?b) =>
	     *   (?a, ?tmp, ?b),
	     *   (?tmp, <rdf:subPropertyOf>, <foo:prop>)
	     */

	    char *q = g_strdup_printf("SELECT count(*) FROM "TBLPREFIX"triples "
		 "WHERE "TBLPREFIX"triples.predicate=%llx AND "TBLPREFIX"triples.object=%llx "
		 "AND "TBLPREFIX"triples.object != "TBLPREFIX"triples.subject",
		 rs_hash(RDFS_SUBPROPERTY), rs_hash(p->p->value));
	    MYSQL_ROW row;
	    MYSQL_RES *res;
	    int subprops;

	    mysql_query(this->db, q);
	    res = mysql_store_result(this->db);
	    if (res) {
		row = mysql_fetch_row(res);
		subprops = atoi(row[0]);
		mysql_free_result(res);
	    } else {
		subprops = 0;
	    }
	    free(q);

	    /* If there is more than (foo, subPropertyOf, foo) */
	    if (subprops > 0) {
		part_t *new_1, *new_2, *new_3;
		pattern_t *new_pa;

		new_1 = new_part(VARIABLE,
				 g_strdup_printf("_tmp%d", tmp_cnt));
		new_2 = new_part(URI, strdup(RDFS_SUBPROPERTY));
		new_3 = new_part(VARIABLE,
				 g_strdup_printf("_tmp%d", tmp_cnt++));
		new_pa = new_pattern(new_1, new_2, p->p, NULL);
		p->p = new_3;
		g_slist_append(query.where, new_pa);
	    }
	}

	if (p->s->type == VARIABLE) {
	    v = constrain_var_type(sym_tbl, p->s->value, ObjURI);
	    if (!v->column) {
		snprintf(tmp, 255, "t%d.subject", i + 1);
		v->column = strdup(tmp);
		v->triple = i + 1;
	    }
	}

	if (p->p->type == VARIABLE) {
	    v = constrain_var_type(sym_tbl, p->p->value, ObjURI);
	    if (!v->column) {
		snprintf(tmp, 255, "t%d.predicate", i + 1);
		v->column = strdup(tmp);
		v->triple = i + 1;
	    }
	}

	if (p->o->type == VARIABLE) {
	    v = constrain_var_type(sym_tbl, p->o->value, ObjAny);
	    if (!v->column) {
		snprintf(tmp, 255, "t%d.object", i + 1);
		v->column = strdup(tmp);
		v->triple = i + 1;
	    }
	}

	if (p->m && p->m->type == VARIABLE) {
	    v = constrain_var_type(sym_tbl, p->m->value, ObjURI);
	    if (!v->column) {
		snprintf(tmp, 255, "t%d.model", i + 1);
		v->column = strdup(tmp);
		v->triple = i + 1;
	    }
	}
    }

    for (li = query.constr; li; li = li->next) {
	constraint_t *c = li->data;

	constraints++;
	if (c->op == EQ_S || c->op == EQ_N || c->op == NE_S ||
	    c->op == NE_N || c->op == GT_S || c->op == GT_N ||
	    c->op == LT_S || c->op == LT_N || c->op == LIKE ||
	    c->op == EQ_RE) {
	    if (c->lval->type == VARIABLE && c->rval->type == URI) {
		constrain_var_type(sym_tbl, c->lval->value, ObjURI);
		require_var_join(sym_tbl, c->lval->value, 1);
	    } else if (c->lval->type == VARIABLE && (c->rval->type ==
		       LITERAL || c->rval->type == INTEGER ||
		       c->rval->type == REGEX)) {
		constrain_var_type(sym_tbl, c->lval->value, ObjLiteral);
		if (c->op == GT_S || c->op == GT_N ||
		    c->op == LT_S || c->op == LT_N ||
                    c->op == LIKE || c->op == EQ_RE) {
		    require_var_join(sym_tbl, c->lval->value, 1);
		}
	    } else if (c->lval->type == VARIABLE &&
		       c->rval->type == VARIABLE) {
		rdql_variable *lv, *rv;

		require_var_join(sym_tbl, c->lval->value, 1);
		require_var_join(sym_tbl, c->rval->value, 1);
		lv = constrain_var_type(sym_tbl, c->lval->value, ObjAny);
		rv = constrain_var_type(sym_tbl, c->rval->value, ObjAny);
		constrain_var_type(sym_tbl, c->lval->value, rv->type);
		constrain_var_type(sym_tbl, c->rval->value, lv->type);
	    } else {
		snprintf(tmp, 255, "unsupported eq operands %d and %d", c->lval->type, c->rval->type);
		return rdql_error(tmp);
	    }
	} else {
	    return rdql_error("unsupported operator");
	}
    }

    selects = new_strvec();
    froms = new_strvec();
    joins = new_strvec();
    wheres = new_strvec();

    cols = calloc(ret_vars * 2 + 1, sizeof(rs_rdql_column));

    for (li = query.select, i = 0; li; li = li->next, i++) {
	rdql_variable *v = g_hash_table_lookup(sym_tbl, li->data);
	v->need_join = 0;
	if (v->type == ObjURI) {
	    cols[col_cnt].variable = (char *) li->data;
	    cols[col_cnt].type = v->type;
	    col_cnt++;
	    strvec_append(selects,
		g_strdup_printf("r%d.uri as %s", i + 1, (char *) li->data));
	    strvec_append(froms, g_strdup_printf(TBLPREFIX"resources r%d", i + 1));
	    v->val_column = g_strdup_printf("r%d.uri", i + 1);
	    v->hash_column = g_strdup_printf("r%d.hash", i + 1);
	    if (strstr(v->column, "object")) {
		strvec_append(wheres,
		    g_strdup_printf("r%d.hash=%s AND t%d.literal=0", i + 1,
				    v->column, v->triple));
	    } else {
		strvec_append(wheres,
		    g_strdup_printf("r%d.hash=%s", i + 1, v->column));
	    }
	} else if (v->type == ObjLiteral) {
	    cols[col_cnt].variable = (char *) li->data;
	    cols[col_cnt].type = v->type;
	    col_cnt++;
	    strvec_append(selects, g_strdup_printf("l%d.literal as %s", i + 1,
				(char *) li->data));
	    strvec_append(froms, g_strdup_printf(TBLPREFIX"literals l%d", i + 1));
	    v->val_column = g_strdup_printf("l%d.literal", i + 1);
	    v->hash_column = g_strdup_printf("l%d.hash", i + 1);
	    if (strstr(v->column, "object")) {
		strvec_append(wheres,
		    g_strdup_printf("l%d.hash=%s AND t%d.literal=1", i + 1,
				    v->column, v->triple));
	    } else {
		strvec_append(wheres,
		    g_strdup_printf("l%d.hash=%s", i + 1, v->column));
	    }
	} else if (v->type == ObjAny) {
	    cols[col_cnt].variable = (char *) li->data;
	    cols[col_cnt].type = ObjURI;
	    col_cnt++;
	    cols[col_cnt].variable = (char *) li->data;
	    cols[col_cnt].type = ObjLiteral;
	    col_cnt++;
	    strvec_append(selects, g_strdup_printf("r%d.uri as _uri_%s, "
					 "l%d.literal as _lit_%s",
					    uid, (char *) li->data,
					    uid, (char *) li->data));
	    strvec_append(joins,
		g_strdup_printf("\nLEFT JOIN "TBLPREFIX"resources r%d ON "
				    "(r%d.hash=%s AND t%d.literal=0)"
				"\nLEFT JOIN "TBLPREFIX"literals l%d ON "
				    "(l%d.hash=%s AND t%d.literal=1)",
				uid, uid, v->column, v->triple,
				uid, uid, v->column, v->triple));
	    uid++;
	}
    }

    /* Iterate over the symbol table and check they are all joined */
    g_hash_table_foreach(sym_tbl, var_walk, froms);

    for (li = query.where, i = 0; li; li = li->next, i++) {
	pattern_t *p = li->data;

	strvec_append(froms, g_strdup_printf(TBLPREFIX"triples t%d", i + 1));

	snprintf(tmp, 255, "t%d.subject", i + 1);
	if (p->s->type == VARIABLE) {
	    rdql_variable *v = g_hash_table_lookup(sym_tbl,
						   p->s->value);
	    /* Don't bother asserting a=a */
	    if (strcmp(v->column, tmp)) {
		strvec_append(wheres,
		    g_strdup_printf("%s=%s", tmp, v->column));
	    }
	} else {
	    strvec_append(wheres, g_strdup_printf("%s=0x%llx", tmp,
						  rs_hash(p->s->value)));
	}

	snprintf(tmp, 255, "t%d.predicate", i + 1);
	if (p->p->type == VARIABLE) {
	    rdql_variable *v = g_hash_table_lookup(sym_tbl,
						   p->p->value);
	    if (strcmp(v->column, tmp)) {
		strvec_append(wheres,
		    g_strdup_printf("%s=%s", tmp, v->column));
	    }
	} else {
	    strvec_append(wheres,
		g_strdup_printf("%s=0x%llx", tmp, rs_hash(p->p->value)));
	}

	snprintf(tmp, 255, "t%d.object", i + 1);
	if (p->o->type == VARIABLE) {
	    rdql_variable *v = g_hash_table_lookup(sym_tbl,
						   p->o->value);
	    if (strcmp(v->column, tmp)) {
		strvec_append(wheres,
		    g_strdup_printf("%s=%s", tmp, v->column));
	    }
	} else {
	    int lit_constr = 0;

	    if (p->o->type == LITERAL || p->o->type == INTEGER) {
		lit_constr = 1;
	    }

	    strvec_append(wheres, g_strdup_printf("%s=0x%llx AND "
		"t%d.literal=%d", tmp, rs_hash(p->o->value), i + 1,
		lit_constr));
	}

	if (p->m) {
	    snprintf(tmp, 255, "t%d.model", i + 1);
	    if (p->m->type == VARIABLE) {
		rdql_variable *v = g_hash_table_lookup(sym_tbl,
						       p->m->value);
		if (strcmp(v->column, tmp)) {
		    strvec_append(wheres,
			g_strdup_printf("%s=%s", tmp, v->column));
		}
	    } else {
		strvec_append(wheres,
		    g_strdup_printf("%s=0x%llx", tmp, rs_hash(p->m->value)));
	    }
	}
    }

    for (li = query.constr; li; li = li->next) {
	constraint_t *c = li->data;
	char *op;

	if (c->op == EQ_S || c->op == EQ_N) {
	    op = "=";
	} else if (c->op == NE_S || c->op == NE_N) {
	    op = "!=";
	} else if (c->op == LT_S || c->op == LT_N) {
	    op = "<";
	} else if (c->op == GT_S || c->op == GT_N) {
	    op = ">";
	} else if (c->op == LIKE) {
	    op = " LIKE ";
	} else if (c->op == EQ_RE) {
	    op = " REGEXP ";
	} else {
	    op = NULL;
	}
	if (op && (c->op == EQ_S || c->op == EQ_N ||
	    c->op == NE_S || c->op == NE_N)) {
	    if (c->lval->type == VARIABLE &&
		(c->rval->type == LITERAL || c->rval->type == URI)) {
		rdql_variable *v = g_hash_table_lookup(sym_tbl,
						       c->lval->value);
		strvec_append(wheres,
		    g_strdup_printf("%s%s0x%llx", v->column,
				    op, rs_hash(c->rval->value)));
	    } else if (c->lval->type == VARIABLE && c->rval->type == INTEGER) {
		rdql_variable *v = g_hash_table_lookup(sym_tbl,
						       c->lval->value);
		strvec_append(wheres,
		    g_strdup_printf("%s=%s AND %s%s%s", v->column,
				    v->hash_column, v->val_column, op,
				    c->rval->value));
	    } else if (c->lval->type == VARIABLE &&
		       c->rval->type == VARIABLE) {
		rdql_variable *lv, *rv;

		lv = g_hash_table_lookup(sym_tbl, c->lval->value);
		rv = g_hash_table_lookup(sym_tbl, c->rval->value);
		strvec_append(wheres,
		    g_strdup_printf("%s%s%s", lv->column, op, rv->column));
	    } else {
		return rdql_error("unsupported operands");
	    }
	} else if (op && (c->op == LT_S || c->op == LT_N ||
		   c->op == GT_S || c->op == GT_N ||
		   c->op == LIKE || c->op == EQ_RE)) {
	    if (c->lval->type == VARIABLE && (c->rval->type == LITERAL)) {
		rdql_variable *v = g_hash_table_lookup(sym_tbl,
						       c->lval->value);
		char *tmp = strdup_escape(this->db, c->rval->value);
		
		strvec_append(wheres,
		    g_strdup_printf("%s=%s AND %s%s\"%s\"", v->column,
				    v->hash_column, v->val_column, op, tmp));
		free(tmp);
	    } else if (c->lval->type == VARIABLE && (c->rval->type == REGEX)) {
		rdql_variable *v = g_hash_table_lookup(sym_tbl,
						       c->lval->value);
		char *tmp = strdup_escape(this->db, c->rval->value);

		strvec_append(wheres,
// XXX need to escape string
		    g_strdup_printf("%s=%s AND %s%s'%s'", v->column,
				    v->hash_column, v->val_column,
				    op, tmp));
	    } else if (c->lval->type == VARIABLE && c->rval->type == INTEGER) {
		rdql_variable *v = g_hash_table_lookup(sym_tbl,
						       c->lval->value);
		strvec_append(wheres,
		    g_strdup_printf("%s=%s AND %s%s%s", v->column,
				    v->hash_column, v->val_column, op,
				    c->rval->value));
	    } else if (c->lval->type == VARIABLE &&
		       c->rval->type == VARIABLE) {
		rdql_variable *lv, *rv;

		lv = g_hash_table_lookup(sym_tbl, c->lval->value);
		rv = g_hash_table_lookup(sym_tbl, c->rval->value);
		strvec_append(wheres,
		    g_strdup_printf("%s=%s AND %s%s%s", lv->column, rv->column, lv->val_column, op, rv->val_column));
	    } else {
		return rdql_error("unsupported operands");
	    }
	} else {
	    return rdql_error("unsupported constraint");
	}
    }

    /* Check to make sure all the variables have been bound */
    for (li = query.select, i = 0; li; li = li->next, i++) {
	rdql_variable *v;

	v = g_hash_table_lookup(sym_tbl, (char *) li->data);
	if (v && !v->column) {
	    return rdql_error(g_strdup_printf("unbound variable ?%s", (char *)li->data));
	}
    }
    g_hash_table_foreach(sym_tbl, st_free, NULL);
    g_hash_table_destroy(sym_tbl);

    select = strvec_join(selects, ", ");
    from = strvec_join(froms, ", ");
    where = strvec_join(wheres, " AND ");
    join = strvec_join(joins, "");
    if (strlen(where) > 0) {
	sql_query = g_strdup_printf("SELECT DISTINCT %s\nFROM %s%s\nWHERE %s%s",
				select, from, join, where, limit ? limit:"");
    } else {
	sql_query = g_strdup_printf("SELECT DISTINCT %s\nFROM %s%s\n%s",
				select, from, join, limit ? limit:"");
    }
    free(select);
    free(from);
    free(join);
    free(where);
    free(limit);

    strvec_free(selects);
    strvec_free(froms);
    strvec_free(joins);
    strvec_free(wheres);

    res = calloc(1, sizeof(rs_rdql_result));
    res->query = calloc(1, sizeof(query_t));
    memcpy(res->query, &query, sizeof(query_t));

    sql_explain = g_strdup_printf("EXPLAIN %s", sql_query);
    if (mysql_query(this->db, sql_explain)) {
	res->error_str =
	    g_strdup_printf("rdql: error in query: %s\nSQL explain error in '%s' %d: %s\n",
			    query_str, sql_query, mysql_errno(this->db),
			    mysql_error(this->db));

	return res;
    }
    res->db_res = mysql_store_result(this->db);
    while (mysql_field_count(this->db) > 5 &&
	   (row = mysql_fetch_row(res->db_res))) {
	if (row[6]) {
	    complexity += atoi(row[6]);
	}
    }
    free(sql_explain);
    mysql_free_result(res->db_res);

    res->sql_query = sql_query;
    if (complexity > max_comp) {
	res->error_str =
	    g_strdup_printf
	    ("Estimated query complexity of %d is greater than limit of %d. Contact your system adminstriator for advice on how to authorise this query.",
	     complexity, max_comp);
	return res;
    }

    if (mysql_query(this->db, sql_query)) {
	res->error_str = g_strdup_printf("SQL error in '%s' %d: %s\n",
					 sql_query, mysql_errno(this->db),
					 mysql_error(this->db));
	fprintf(stderr, res->error_str);
	fprintf(stderr, "Query: %s\n", sql_query);

	return res;
    }

    res->db_res = mysql_store_result(this->db);
    res->columns = col_cnt;
    res->row = cols;

    return res;
}

rs_rdql_column *rs_rdql_next_row(rs_rdql_result * res)
{
    MYSQL_ROW row;
    unsigned int i;

    if (!res || !(res->db_res)) {
	return NULL;
    }

    row = mysql_fetch_row((MYSQL_RES *) (res->db_res));
    if (!row) {
	return NULL;
    }

    for (i = 0; i < res->columns; i++) {
	res->row[i].value = row[i];
    }

    return res->row;
}

void rs_rdql_free_result(rs_rdql_result * res)
{
    if (res) {
	mysql_free_result((MYSQL_RES *) res->db_res);
	free(res->sql_query);
	free(res->row);
	query_t_free(res->query);
	free(res);
    }
}

rs_rdql_result *rdql_error(const char *fmt, ...)
{
    va_list ap;
    rs_rdql_result *res;

    res = calloc(1, sizeof(rs_rdql_result));
    va_start(ap, fmt);
    res->error_str = g_strdup_vprintf(fmt, ap);

    return res;
}

void yyerror(char *msg)
{
    y_err_str = g_strdup_printf("Parse error: %s, line %d\n", msg, lineno);
}

/* This function walks the symbol table and binds any literal variables that
 * are compared but not returned to a literal table so the SQL comparators can
 * be used */
void var_walk(gpointer key, gpointer value, gpointer user_data)
{
    strvec_t *froms = user_data;
    rdql_variable *v = value;

    if (v->need_join) {
	if (!v->val_column) {
	    if (v->type == ObjLiteral) {
		const int num = strvec_length(froms) + 1;
		//v->column = g_strdup_printf(TBLPREFIX"literals l%d", num);
		//strvec_append(froms, v->column);
		strvec_append(froms, g_strdup_printf(TBLPREFIX"literals l%d", num));
		v->val_column = g_strdup_printf("l%d.literal", num);
		v->column = g_strdup_printf("l%d.hash", num);
		v->hash_column = g_strdup_printf("l%d.hash", num);
	    } else if (v->type == ObjURI) {
		const int num = strvec_length(froms) + 1;
		v->column = g_strdup_printf(TBLPREFIX"resources r%d", num);
		strvec_append(froms, v->column);
		v->val_column = g_strdup_printf("r%d.uri", num);
		v->hash_column = g_strdup_printf("r%d.hash", num);
		//v->column = v->hash_column;
	    } else if (v->type != ObjAny) {
		/* type is known, but unsupported */
		fprintf(stderr, "Unhandled variable '%s' type %d at %s:%d\n", v->name, v->type, __FILE__, __LINE__);
	    }
	}
    }
}

char *strdup_escape(MYSQL *db, const char *s)
{
    char *ret;
    int len;

    len = strlen(s);
    ret = malloc(len*2 + 1);
    mysql_real_escape_string(db, ret, s, len);

    return ret;
}

/* vi:set ts=8 sts=4 sw=4: */
