#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "rdfsql/rdql.h"
#include "rdfsql/rdql_datatypes.h"
#include "rdql_parser.h"

rdql_variable *new_rdql_variable(const char *name)
{
	rdql_variable *v = malloc(sizeof(rdql_variable));

	v->name = name;
	v->value = NULL;
	v->column = NULL;
	v->val_column = NULL;
	v->hash_column = NULL;
	v->type = ObjAny;
	v->need_join = 0;

	return v;
}

char *rdql_variable_type_pp(rdql_variable *v)
{
	switch (v->type) {
		case ObjURI:
			return "uri";
			break;
		case ObjLiteral:
			return "literal";
			break;
		default:
			return "unknown";
			break;
	}

	return NULL;
}

rdql_variable *constrain_var_type(GHashTable *st, const char *name,
		rs_obj_type type)
{
	rdql_variable *v = g_hash_table_lookup(st, name);

	if (v && v->type == ObjAny) {
		 v->type = type;
	} else if (v && type != ObjAny && v->type != type) {
		rdql_error("confilicting types for ?%s", name);
	} else if (!v) {
		v = new_rdql_variable(name);
		v->type = type;
		g_hash_table_insert(st, (char *)name, v);
	}

	return v;
}

rdql_variable *require_var_join(GHashTable *st, const char *name,
		int need_join)
{
	rdql_variable *v = g_hash_table_lookup(st, name);

	if (!v) {
		v = new_rdql_variable(name);
		v->type = ObjAny;
		g_hash_table_insert(st, (char *)name, v);
	}
	v->need_join = need_join;

	return v;
}

pattern_t *new_pattern(part_t *s, part_t *p, part_t *o, part_t *m)
{
	pattern_t *pa = malloc(sizeof(pattern_t));
	pa->s = s;
	pa->p = p;
	pa->o = o;
	pa->m = m;

	return pa;
}

void pattern_free(pattern_t *p)
{
	free(p->s);
	free(p->p);
	free(p->o);
	free(p->m);
	free(p);
}

void expand_pattern(pattern_t *p, GSList *prefs)
{
	expand_part(p->s, prefs);
	expand_part(p->p, prefs);
	expand_part(p->o, prefs);
	expand_part(p->m, prefs);
}

part_t *new_part(int type, const char *value)
{
	part_t *p = malloc(sizeof(part_t));
	p->type = type;
	p->value = value;

	return p;
}

void expand_part(part_t *p, GSList *prefs)
{
	GSList *li;
	char *tmp;
	unsigned int len;

	if (!p || p->type != URI) {
		return;
	}
	len = strlen(p->value);

	for (li = prefs; li; li = li->next) {
		const prefix_t *pre = (prefix_t *)(li->data);

		if (len > pre->prefix_len &&
		    p->value[pre->prefix_len] == ':' &&
		    !strncmp(pre->prefix, p->value, strlen(pre->prefix))) {
			tmp = (char *)p->value;
			p->value = g_strconcat(pre->uri, p->value +
					pre->prefix_len + 1, NULL);
			free(tmp);
			return;
		}
	}
}

prefix_t *new_prefix(const char *prefix, const char *uri)
{
	prefix_t *p = malloc(sizeof(prefix_t));
	p->prefix = prefix;
	p->prefix_len = strlen(prefix);
	p->uri = uri;

	return p;
}

constraint_t *new_constraint(part_t *lval, unsigned int op, part_t *rval)
{
	constraint_t *c = malloc(sizeof(constraint_t));

	c->lval = lval;
	c->op = op;
	c->rval = rval;

	return c;
}

strvec_t *new_strvec()
{
	strvec_t *sv = malloc(sizeof(strvec_t));
	sv->length = 0;
	sv->size = 16;
	sv->vec = calloc(sizeof(char *), sv->size);

	return sv;
}

void strvec_append(strvec_t *sv, const char *s)
{
	sv->vec[sv->length] = (char *)s;
	sv->length++;
	if (sv->length >= sv->size) {
		int i;

		sv->size *= 2;
		sv->vec = realloc(sv->vec, sizeof(char *) * sv->size);
		for (i = sv->length; i < sv->size; i++) {
			sv->vec[i] = NULL;
		}
	}
}

int strvec_length(strvec_t *sv)
{
	return sv->length;
}

char *strvec_join(strvec_t *sv, const char *inter)
{
	return g_strjoinv(inter, sv->vec);
}

void strvec_free(strvec_t *sv)
{
	int i;

	for (i=0; i<sv->length; i++) {
		free(sv->vec[i]);
	}
	free(sv->vec);
	free(sv);
}

static void list_free_data(GSList *l)
{
	GSList *it;

	for (it=l; it; it=it->next) {
		free(it->data);
	}
}

static void list_free_where_data(GSList *l)
{
	GSList *it;

	for (it=l; it; it=it->next) {
		pattern_free(it->data);
	}
}

void query_t_free(query_t *q)
{
	list_free_data(q->select);
	g_slist_free(q->select);
	list_free_data(q->from);
	g_slist_free(q->from);
	list_free_where_data(q->where);
	g_slist_free(q->where);
	list_free_data(q->prefixes);
	g_slist_free(q->prefixes);
	list_free_data(q->constr);
	g_slist_free(q->constr);
}
