/* Parser */

%{

#include <string.h>

#include "lex.h"
#include "rdfsql/rdql.h"

#define YY_NO_UNPUT
#define YYDEBUG 1
#define YYERROR_VERBOSE

query_t q_data;

void yyerror (char *msg);

%}

/* The structure for passing value between lexer and parser */
%union {
   char         *str;
   pattern_t    *pattern;
   constraint_t *constraint;
   part_t       *part;
   int		 op;
   int		 integer;
   GSList       *list;
}

/* structural symbols */
%token OPEN_PAR CLOSE_PAR COMMA

/* syntax symbols */
%token SELECT FROM WHERE USING FOR AND LIMIT OFFSET

/* values */
%token URI LITERAL VARIABLE SYMBOL INTEGER REGEX

/* operators */
%token EQ_S NE_S GT_S LT_S EQ_N NE_N GT_N LT_N LIKE EQ_RE

/* end of query */
%token EOQ

%%

query		: SELECT var_list from_spec WHERE pattern_list constraint_spec prefix_spec limit_spec offset_spec eoq
		  { q_data.select = $2.list;
		    q_data.from = $3.list;
		    q_data.where = $5.list;
		    q_data.constr = $6.list;
		    q_data.prefixes = $7.list;
		    q_data.limit = $8.integer;
		    q_data.offset = $9.integer;
		  }
		;

from_spec	: FROM uri_list
		  { $$.list = $2.list; }
		| /* empty */
		  { $$.list = NULL; }
		;

var_list	: VARIABLE COMMA var_list
		  { $$.list = g_slist_prepend($3.list, $1.str); }
		| VARIABLE var_list
		  { $$.list = g_slist_prepend($2.list, $1.str); }
		| VARIABLE
		  { $$.list = g_slist_alloc();
		    $$.list->data = $1.str; }
		;

uri_list	: URI COMMA uri_list
		  { $$.list = g_slist_prepend($3.list, $1.str); }
		| URI uri_list
		  { $$.list = g_slist_prepend($2.list, $1.str); }
		| URI
		  { $$.list = g_slist_alloc();
		    $$.list->data = $1.str; }
		;

pattern_list	: pattern COMMA pattern_list
		  { $$.list = g_slist_prepend($3.list, $1.pattern); }
		| pattern pattern_list
		  { $$.list = g_slist_prepend($2.list, $1.pattern); }
		| pattern
		  { $$.list = g_slist_alloc();
		    $$.list->data = $1.pattern; }
		;

pattern		: OPEN_PAR part COMMA part COMMA part CLOSE_PAR
		  { $$.pattern = new_pattern($2.part, $4.part, $6.part, NULL); }
		| OPEN_PAR part part part CLOSE_PAR
		  { $$.pattern = new_pattern($2.part, $3.part, $4.part, NULL); }
		| OPEN_PAR part COMMA part COMMA part COMMA part CLOSE_PAR
		  { $$.pattern = new_pattern($2.part, $4.part, $6.part, $8.part); }
		| OPEN_PAR part part part part CLOSE_PAR
		  { $$.pattern = new_pattern($2.part, $3.part, $4.part, $5.part); }
		;

part		: URI
		  { $$.part = new_part(URI, $1.str); }
		| LITERAL
		  { $$.part = new_part(LITERAL, $1.str); }
		| VARIABLE
		  { $$.part = new_part(VARIABLE, $1.str); }
		| INTEGER
		  { char *canon = g_strdup_printf("%d", atoi($1.str));
		    $$.part = new_part(INTEGER, canon); }
		;

prefix_spec	: USING prefix_list
		  { $$.list = $2.list; }
		| /* empty */
		  { $$.list = NULL; }
		;

prefix_list	: SYMBOL FOR URI COMMA prefix_list
		  { $$.list = g_slist_prepend($5.list, new_prefix($1.str, $3.str)); }
		| SYMBOL FOR URI prefix_list
		  { $$.list = g_slist_prepend($4.list, new_prefix($1.str, $3.str)); }
		| SYMBOL FOR URI
		  { $$.list = g_slist_alloc();
		    $$.list->data = new_prefix($1.str, $3.str); }
		;

constraint_spec	: AND constraint constraint_spec
		  { $$.list = g_slist_prepend($3.list, $2.constraint); }
		| /* empty */
		  { $$.list = NULL; }
		;

constraint	: part operator part
		  { $$.constraint = new_constraint($1.part, $2.op, $3.part); }
		| VARIABLE EQ_RE REGEX
		  { $$.constraint = new_constraint(new_part(VARIABLE, $1.str),
						   EQ_RE,
						   new_part(REGEX, $3.str)); }
		;

operator	: EQ_S
		  { $$.op = EQ_S; }
		| NE_S
		  { $$.op = NE_S; }
		| GT_S
		  { $$.op = GT_S; }
		| LT_S
		  { $$.op = LT_S; }
		| EQ_N
		  { $$.op = EQ_N; }
		| NE_N
		  { $$.op = NE_N; }
		| GT_N
		  { $$.op = GT_N; }
		| LT_N
		  { $$.op = LT_N; }
		| LIKE
		  { $$.op = LIKE; }
		| EQ_RE
		  { $$.op = EQ_RE; }

limit_spec	: LIMIT INTEGER
		  { $$.integer = atoi($2.str); }
		| /* empty */
		  { $$.integer = 0; }
		;

offset_spec	: OFFSET INTEGER
		  { $$.integer = atoi($2.str); }
		| /* empty */
		  { $$.integer = 0; }
		;

eoq		: EOQ
		| /* empty */
		;

%%
