/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: tstore_import.c,v 1.16 2004/06/21 13:23:19 theno23 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <mysql.h>
#include <sys/types.h>
#include <getopt.h>
#include <glib.h>

#include "rdfsql/rdfsql.h"
#include "rst_parser.h"

#define INT_NS "http://triplestore.aktors.org/ontology/#"

#define MEMBER_PREFIX "http://www.w3.org/1999/02/22-rdf-syntax-ns#_"
#define MEMBER_PROP   "http://www.w3.org/2000/01/rdf-schema#member"

#define INTEGER_DT    "http://www.w3.org/2001/XMLSchema#integer"

static RDFSQL *db;

static char *progname;

static int verbosity = 0;

int main(int argc, char *argv[])
{
    char uri[512];
    char *model = NULL;
    unsigned int i;
    int lock = 0;
    int show_help = 0, flush = 0;
    char *optstring = RST_OPTSTR "m:flv";
    char *syntax = "auto";
    rs_syntax format = RS_AUTO;
    int c, opt_index = 0;
    static struct option long_options[] = {
	RST_LONGOPTS,
	{"model", 1, 0, 'm'},
	{"lock", 0, 0, 'l'},
	{"flush", 0, 0, 'f'},
	{"verbose", 0, 0, 'v'},
	{"syntax", 1, 0, 's'},
	{0, 0, 0, 0}
    };

    while ((c = getopt_long(argc, argv, optstring, long_options,
			&opt_index)) != -1) {
	if (!rst_handle_option(c, optarg)) {
	    switch (c) {
	    case 'h':
		show_help = 1;
		break;
	    case 'f':
		flush = 1;
		break;
	    case 'l':
		lock = 1;
		break;
	    case 'm':
		model = optarg;
		break;
	    case 's':
		syntax = optarg;
		break;
	    case 'v':
		verbosity++;
		break;
	    default:
		fprintf(stderr, "Unknown option '%c'\n", c);
		show_help = 1;
		break;
	    }
	}
    }

    if (show_help) {
	printf("Usage: %s " RST_OPTIONS
	       " [-f] [-l] [-m model-uri] <rdf-file>...\n", argv[0]);
	printf(RST_HELP);
	printf("   -f, --flush     flush model before asserting\n");
	printf("   -l, --lock      write lock tables when inserting (faster, prevents reads)\n");
	printf("   -m, --model     specify RDF model URI\n");
	printf("   -s, --syntax    specify the RDF syntax, one of:\n");
	printf("                      xml, ntriples, turtle, auto (default)");
	exit(1);
    }

    if (!strcasecmp(syntax, "xml")) {
	format = RS_XML;
    } else if (!strcasecmp(syntax, "ntriples")) {
	format = RS_NTRIPLES;
    } else if (!strcasecmp(syntax, "turtle")) {
	format = RS_TURTLE;
    } else {
	format = RS_AUTO;
    }

    progname = argv[0];

    if (!model) {
	if (strncmp("http:", argv[optind], 5) && strncmp("file:",
		    argv[optind], 5)) {
	    snprintf(uri, 300, "file:%s", argv[optind]);
	} else {
	    snprintf(uri, 300, "%s", argv[optind]);
	}
	model = uri;
    }

    if (verbosity > 0) printf("using model <%s>\n", model);
    db = rst_connect(model);

    if (lock) {
	if (mysql_query(db->db, "LOCK TABLES "TBLPREFIX"triples WRITE, "TBLPREFIX"resources WRITE, "
			  TBLPREFIX"literals WRITE, "TBLPREFIX"models WRITE,"TBLPREFIX"cache_state WRITE")) {
	    fprintf(stderr, "failed to lock tables (%d): %s\n",
		    mysql_errno(db->db), mysql_error(db->db));
	    lock = 0;
	}

    }

    if (flush) {
	if (verbosity > 0) {
	    printf("flushing %s (%lld)\n", model, db->model);
	}
	rs_flush(db);
    }

    for (i = optind; i < argc; i++) {
	if (strncmp("http:", argv[i], 5) && strncmp("file:", argv[i], 5)) {
	    snprintf(uri, 511, "file:%s", argv[i]);
	} else {
	    snprintf(uri, 511, "%s", argv[i]);
	}
	if (verbosity > 0) printf("Reading '%s'\n", uri);

	rs_import2(db, uri, model, verbosity, stderr, format);
    }

    if (lock) {
        if (mysql_query(db->db, "UNLOCK TABLES")) {
            fprintf(stderr, "Failed to unlock tables (%d): %s\n",
                    mysql_errno(db->db), mysql_error(db->db));
        }
    }

    rs_update_caches(db, verbosity);

    return 0;

}

/* vi:set ts=8 sts=4 sw=4: */
