/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: tstore_rdql.c,v 1.13 2005/03/22 19:21:33 theno23 Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "rst_parser.h"
#include "rdfsql/rdfsql.h"

void interactive(RDFSQL *db);
void batch(RDFSQL *db);
int run_query(RDFSQL *db, const char *query);
const char *escape_str(const char *in);
void cleanup();
void set_format(const char *fmt);
int handle_commands(const char *line);

static int perf_info = 0;
static int max_comp = 5000;

#define FORMATS 2
static const char * const formats[FORMATS] = {"tabular", "list"};
static int format = 0;

char *history_fn = NULL;

static char *find_terminator(const char *line);

int main(int argc, char *argv[])
{
    RDFSQL *db;
    int help = 0;
    char *optstring = RST_OPTSTR "m:vf:";
    int c, opt_index = 0;
    static struct option long_options[] = {
	RST_LONGOPTS,
	{ "max-comp", 1, 0, 'm' },
	{ "load"    , 0, 0, 'l' },
	{ "verbose" , 1, 0, 'v' },
	{ "format"  , 1, 0, 'f' },
	{ 0, 0, 0, 0 }
    };

    while ((c = getopt_long (argc, argv, optstring, long_options, &opt_index)) != -1) {
	if (!rst_handle_option(c, optarg)) {
	    switch (c) {
	    case 'h':
		help++;
		break;
	    case 'm':
		max_comp = atoi(optarg);
		break;
	    case 'v':
		perf_info = 1;
		break;
	    case 'f':
		set_format(optarg);
		break;
	    default:
		help++;
		break;
	    }
	}
    }


    if (help || argc - optind > 1) {
	fprintf(stderr, "Usage: %s "RST_OPTIONS" [-m <comp>] [-f <format>] [-v] [query]\n" RST_HELP
"   -m, --max-comp  maximum query complexity (default %d)\n"
"   -f, --format=?  set output format %s or %s (default %s)\n"
"   -v, --verbose   show performance information and translation\n", argv[0], max_comp, formats[0], formats[1], formats[format]);
	return 1;
    }

    db = rst_connect(NULL);
    if (!db) {
	fprintf(stderr, "Could not connect to database, is mysql running "
			"and have you set up the database tables as described "
			"in the README?\n");
	return 1;
    }

    if (argc - optind == 0 || !strcmp(argv[optind], "-")) {
	if (isatty(0)) {
	    interactive(db);
	} else {
	    batch(db);
	}
    }

    run_query(db, argv[optind]);

    rs_disconnect(db);

    return 0;
}

void set_format(const char *fmt)
{
    int found = 0;
    int i;

    for (i = 0; i < FORMATS; i++) {
	if (!strcmp(fmt, formats[i])) {
	    format = i;
	    found = 1;
	}
    }
    if (!found) {
	fprintf(stderr, "error: cannot match format '%s'\n", fmt);
    }
}

#define MAX_Q_SIZE 16384

void interactive(RDFSQL *db)
{
    char *line, *last_line = NULL;
    char curr_line[MAX_Q_SIZE];
    const char *base_pr = "rdql> ";
    const char *cont_pr = "    > ";
    FILE *history;

    atexit(cleanup);

    history_fn = g_strdup_printf("%s/.rdql_history", getenv("HOME"));
    history = fopen(history_fn, "r");
    while (history && !feof(history)) {
	fgets(curr_line, sizeof(char) * MAX_Q_SIZE, history);
	if (feof(history)) {
	    break;
	}
	if (strlen(curr_line) > 0) {
	    curr_line[strlen(curr_line) - 1] = '\0';
	}
	add_history(strdup(curr_line));
    }

    curr_line[0] = '\0';

    while ((line = readline(curr_line[0] ? cont_pr : base_pr))) {
	if (!last_line || (line && *line && strcmp(line, last_line))) {
	    add_history(line);
	}
	free(last_line);
	last_line = strdup(line);

	while (line) {
	    //char *pos = strchr(line, ';');
	    char *pos = find_terminator(line);
	    if (pos) {
		*pos = '\0';
		strncat(curr_line, line, MAX_Q_SIZE);
		if (!handle_commands(curr_line)) {
		    run_query(db, curr_line);
		}
		curr_line[0] = '\0';
		find_terminator(NULL);
		line = pos + 1;
	    } else {
		break;
	    }
	}
	if (line && *line) {
	    strncat(curr_line, line, MAX_Q_SIZE);
	    strncat(curr_line, " ", MAX_Q_SIZE);
	}
    }

    printf("quit;\n");
    exit(0);
}

static char *find_terminator(const char *line)
{
    static int in_single = 0;
    static int in_double = 0;

    if (!line) {
	in_single = 0;
	in_double = 0;
    } else {
	char *pos;

	for (pos = (char *)line; *pos; pos++) {
	    if (in_single) {
		if (*pos == '\'') {
		    in_single = 0;
		}
	    } else if (in_double) {
		if (*pos == '"') {
		    in_double = 0;
		}
	    } else {
		if (*pos == '\'') {
		    in_single = 1;
		} else if (*pos == '"') {
		    in_double = 1;
		} else if (*pos == ';') {
		    return pos;
		}
	    }
	}
    }

    return NULL;
}

int handle_commands(const char *line)
{
    char s_name[32], s_val[32];
    char *fmts[] = {
	"SET %31s %31s",
	"set %31s %31s",
	NULL
    };
    int f, n = 0;

    if (!strncmp(line, "QUIT", 4) ||
	!strncmp(line, "quit", 4) ||
	!strncmp(line, "EXIT", 4) ||
	!strncmp(line, "exit", 4)) {
	exit(0);
    }

    s_name[0] = '\0';
    for (f = 0; fmts[f] && n != 2; f++) {
	n = sscanf(line, fmts[f], s_name, s_val);
    }

    if (!s_name[0]) {
	return 0;
    }

    if (!strcmp(s_name, "max_comp")) {
	max_comp = atoi(s_val);
    } else if (!strcmp(s_name, "format")) {
	set_format(s_val);
    } else {
	fprintf(stderr, "error: unknown var '%s'\n", s_name);
    }
    return 1;
}

void batch(RDFSQL *db)
{
    char line_buf[MAX_Q_SIZE], curr_line[MAX_Q_SIZE];
    char *line = NULL;

    curr_line[0] = '\0';

    while (fgets(line_buf, MAX_Q_SIZE, stdin)) {
	if (strlen(line_buf) == 0) {
	    continue;
	}
	line_buf[strlen(line_buf) - 1] = '\0';
	line = line_buf;
	while (line) {
	    char *pos = find_terminator(line);
	    if (pos) {
		*pos = '\0';
		strncat(curr_line, line, MAX_Q_SIZE);
		if (!handle_commands(curr_line)) {
		    run_query(db, curr_line);
		    printf("EOR\n");
		    fflush(stdout);
		}
		curr_line[0] = '\0';
		line = pos + 1;
		find_terminator(NULL);
	    } else {
		break;
	    }
	}
	if (*line) {
	    strncat(curr_line, line, MAX_Q_SIZE);
	    strncat(curr_line, " ", MAX_Q_SIZE);
	}
    }

    exit(0);
}

int run_query(RDFSQL *db, const char *query)
{
    rs_rdql_result *rdql_res;
    rs_rdql_column *row;
    unsigned int i;
    char *last_var = "";

fprintf(stderr, "Query: '%s'\n", query);
    rdql_res = rs_rdql_query(db, query, max_comp);

    if (rdql_res->error_str) {
	printf("%s\n", rdql_res->error_str);
	if (perf_info) {
	    printf ("SQL: %s\n", rdql_res->sql_query);
	}
	return 1;
    }

    if (perf_info) {
	printf ("SQL: %s\n", rdql_res->sql_query);
    }

    if (format == 0) {
	for (i = 0; i < rdql_res->columns; i++) {
	    if (strcmp(rdql_res->row[i].variable, last_var)) {
		printf("?%s\t", rdql_res->row[i].variable);
	    }
	    last_var = (char *)rdql_res->row[i].variable;
	}
	printf("\n");

	while ((row = rs_rdql_next_row(rdql_res))) {
	    for (i = 0; i < rdql_res->columns; i++) {
		if (row[i].value) {
		    if (row[i].type == ObjURI) {
			printf("<%s>\t", row[i].value);
		    } else {
			printf("\"%s\"\t", escape_str(row[i].value));
		    }
		}
	    }
	    printf("\n");
	}
    } else {
	while ((row = rs_rdql_next_row(rdql_res))) {
	    for (i = 0; i < rdql_res->columns; i++) {
		if (row[i].value) {
		    if (row[i].type == ObjURI) {
			printf("?%s\t<%s>\n", row[i].variable, row[i].value);
		    } else {
			printf("?%s\t\"%s\"\n", row[i].variable,
			       escape_str(row[i].value));
		    }
		}
	    }
	    printf("\n");
	}
    }
    rs_rdql_free_result(rdql_res);

    return 0;
}

/* XXX this is not thredsafe */

const char *escape_str(const char *in)
{
    static int length = 0;
    static char *buffer = NULL;
    int newlen;
    char *ptr;
    char *outp;

    newlen = strlen(in) * 2 + 1;
    if (newlen > length) {
	buffer = realloc(buffer, newlen);
    }

    for (ptr = (char *)in, outp = buffer; *ptr; ptr++) {
	switch (*ptr) {
	    case '"':
		*outp++ = '\\';
		*outp++ = '"';
		break;
	    case '\n':
		*outp++ = '\\';
		*outp++ = 'n';
		break;
	    case '\t':
		*outp++ = '\\';
		*outp++ = 't';
		break;
	    case '\\':
		*outp++ = '\\';
		*outp++ = '\\';
		break;
	    default:
		*outp++ = *ptr;
		break;
	}
    }
    *outp = '\0';

    return buffer;
}

/* maximum numer of lines that will be remebered */
#define MAX_HIST 32

void cleanup()
{
    HIST_ENTRY *h;
    int count = 0, i;
    FILE *history;
    char *hist[MAX_HIST];

    if (history_fn) {
	history = fopen(history_fn, "w");
	if (!history) {
	    perror("warning: cannot write history file\n");
	    return;
	}
	while ((h = previous_history()) && count < MAX_HIST) {
	    if (h->line) {
		hist[count++] = h->line;
	    }
	}
	for (i = count-1; i >= 0; i--) {
	    fprintf(history, "%s\n", hist[i]);
	}
    }
}

/* vi:set ts=8 sts=4 sw=4: */
