/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: tstore_setup.c,v 1.6 2004/05/23 13:26:29 theno23 Exp $
 */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <mysql.h>
#include <mysqld_error.h>
#include <errmsg.h>
#include <getopt.h>
#include <locale.h>
#include <readline/readline.h>

#include "rdfsql/rdfsql.h"
#include "rst_parser.h"

void create_db(MYSQL *mdb);
void create_user(MYSQL *mdb);
void create_tables(MYSQL *mdb);
void erase_db(MYSQL *mdb);

static int verbosity = 0;
static int db_exists = 0;

#define LOG(l, a...) { if (verbosity >= l) printf(a); }

int main(int argc, char *argv[])
{
    MYSQL mdb;
    char *optstring = RST_OPTSTR "vse";
    int help = 0;
    int show_uris = 0;
    int erase = 0;
    int c, opt_index = 0;

    static struct option long_options[] = {
        RST_LONGOPTS,
        { "verbose", 0, 0, 'v' },
        { "show", 0, 0, 's' },
        { "erase", 0, 0, 'e' },
        { 0, 0, 0, 0 }
    };

    setlocale(LC_ALL, NULL);

    while ((c = getopt_long (argc, argv, optstring, long_options, &opt_index)) != -1) {
        if (!rst_handle_option(c, optarg)) {
            if (c == 'v') {
                verbosity++;
            } else if (c == 's') {
                show_uris = 1;
            } else if (c == 'e') {
                erase = 1;
            } else {
                help++;
            }
        }
    }

    if (help) {
        printf("Usage: %s "RST_OPTIONS" [-v] \n", argv[0]);
        printf(RST_HELP);
        printf("   -v, --verbose   increase verbosity\n");
        printf("   -s, --show      show uris\n");
        printf("This command creates new 3store users and/or KBs.\n");
        return 1;
    }

    /* Check if user exists and can access the KB */
    mysql_init(&mdb);

    if (erase) {
	if (!mysql_real_connect(&mdb, rst_get_host(), rst_get_user(),
			    rst_get_pw(), rst_get_db(), 0, NULL, 0)) {
	    fprintf(stderr, "Failed to connect to database: Error %d: %s\n", mysql_errno(&mdb), mysql_error(&mdb));
	    exit(1);
	}
	erase_db(&mdb);
	LOG(1, "database erased\n");
    }

    while (!mysql_real_connect(&mdb, rst_get_host(), rst_get_user(),
			    rst_get_pw(), rst_get_db(), 0, NULL, 0)) {
	/* User cant access db, or something else is wrong */
	int err = mysql_errno(&mdb);

	switch(err) {
	case CR_CONN_HOST_ERROR:
	case CR_CONNECTION_ERROR:
	    fprintf(stderr, "Could not connect to mysql server - ensure server is running\n");
	    exit(1);
	    break;
	case CR_UNKNOWN_HOST:
	    fprintf(stderr, "Unknown server host name.\n");
	    exit(1);
	    break;
	case CR_VERSION_ERROR:
	    fprintf(stderr, "Protocol mismatch - this verison of 3store was not compiled against the correct mysql version.\n");
	    exit(1);
	    break;
	case ER_DBACCESS_DENIED_ERROR:
	case ER_ACCESS_DENIED_ERROR:
	    /* These are expeted errors if the system has never been set up */
	    create_db(&mdb);
	    create_user(&mdb);
	    break;
	case ER_BAD_DB_ERROR:
	    /* This error is expected if the user exists but the database
             * doesnt, eg. its been dropped */
	    create_db(&mdb);
	    break;
	default:
	    fprintf(stderr, "Failed to connect to database: Error %d: %s\n", err, mysql_error(&mdb));
	    exit(1);
	    break;
	}
    }

    if (!db_exists) create_tables(&mdb);

    return 0;
}

void create_db(MYSQL *mdb)
{
    char *tmp;

    LOG(1, "Database not created, attempting to create.\n");
    if (!mysql_real_connect(mdb, rst_get_host(), "root", "", "mysql",
			    0, NULL, 0)) {
	int i;
	char *pw;
	MYSQL *res = NULL;

	LOG(1, "MySQL root password required for 'mysql' database.\n");
	for (i=0; i<3 && !res; i++) {
	    pw = readline("MySQL root password: ");
	    res = mysql_real_connect(mdb, rst_get_host(), "root", pw, "mysql",
			    0, NULL, 0);
	    /* due dilligence - dont really want the root pw kicking around */
	    memset(pw, 0, strlen(pw));
	    free(pw);
	}
	if (!res) {
	    fprintf(stderr, "Too many password attempts, aborting.\n");
	    exit(1);
	}
    }
    LOG(1, "Logged in as root\n");

    tmp = g_strdup_printf("CREATE DATABASE %s", rst_get_db());
    if (mysql_query(mdb, tmp)) {
	if (mysql_errno(mdb) == ER_DB_CREATE_EXISTS) {
	    LOG(1, "Database allready exists. Good.\n");
	    db_exists = 1;
	} else {
	    fprintf(stderr, "MySQL error %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
	    exit(1);
	}
    }
    free(tmp);
}

void erase_db(MYSQL *mdb)
{
    char *tmp;

    if (!mysql_real_connect(mdb, rst_get_host(), "root", "", "mysql",
			    0, NULL, 0)) {
	int i;
	char *pw;
	MYSQL *res = NULL;

	LOG(1, "MySQL root password required for 'mysql' database.\n");
	for (i=0; i<3 && !res; i++) {
	    pw = readline("MySQL root password: ");
	    res = mysql_real_connect(mdb, rst_get_host(), "root", pw, "mysql",
			    0, NULL, 0);
	    /* due dilligence - dont really want the root pw kicking around */
	    memset(pw, 0, strlen(pw));
	    free(pw);
	}
	if (!res) {
	    fprintf(stderr, "Too many password attempts, aborting.\n");
	    exit(1);
	}
    }
    LOG(1, "Logged in as root\n");

    tmp = g_strdup_printf("DROP DATABASE %s", rst_get_db());
    if (mysql_query(mdb, tmp)) {
	fprintf(stderr, "MySQL error %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
	exit(1);
    }
    free(tmp);
}

void create_user(MYSQL *mdb)
{
    char *tmp;

    LOG(1, "User not created, or not allowed to access specified KB, attempting\nto create user.\n");
    tmp = g_strdup_printf("GRANT ALL PRIVILEGES ON %s.* TO %s@localhost IDENTIFIED BY '%s'", rst_get_db(), rst_get_user(), rst_get_pw());
    if (mysql_query(mdb, tmp)) {
	fprintf(stderr, "MySQL error %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
	exit(1);
    }
    free(tmp);
    LOG(1, "Created user and given permissions.\n");
}

void create_tables(MYSQL *mdb)
{
    FILE *tables;
    char tmp[4096], *c;
    int pos = 0;

    LOG(1, "Creating tables.\n");
    snprintf(tmp, 4095, "USE %s", rst_get_db());
    if (mysql_query(mdb, tmp)) {
       fprintf(stderr, "MySQL error attempting to swtich to new db: %d: %s\n",
	       mysql_errno(mdb), mysql_error(mdb));
       exit(1);
    }
    if (!(tables = fopen(PKG_DATA_DIR "/tables.sql", "r"))) {
	perror("Cannot open table description file");
	fprintf(stderr, "Cannot create KB tables. Have you run 'make install'?\n");
	exit(1);
    }

    while (fgets(tmp+pos, sizeof(tmp) - pos, tables)) {
	pos = strlen(tmp);
	if (strchr(tmp, ';')) {
	    for (c=tmp; *c; c++) {
		if (isspace(*c)) {
		    *c = ' ';
		}
		if (*c == ';') {
		    *c = '\0';
		}
	    }
	    LOG(1, "SQL: ");
	    LOG(1, tmp);
	    LOG(1, "\n");
	    if (mysql_query(mdb, tmp)) {
		if (mysql_errno(mdb) == ER_TABLE_EXISTS_ERROR) {
		    LOG(1, "%s: looks like table is allready created.\n",
			   mysql_error(mdb));
		} else {
		    fprintf(stderr, "MySQL error %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
		    exit(1);
		}
	    }
	    pos = 0;
	    *tmp = '\0';
	}
    }
}

/* vi:set ts=8 sts=4 sw=4: */
