/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: tstore_test.c,v 1.6 2004/10/18 14:22:15 theno23 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <mysql.h>
#include <getopt.h>
#include <sys/time.h>

#include "rdfsql/rdfsql.h"
#include "rst_parser.h"

#define FAIL(fmt, ...) \
    fprintf(stderr, "failed: " fmt "\n", ## __VA_ARGS__); exit(2)

int res_test(rs_result *res, const char *s, const char *p, const char *o);

typedef enum {
    t_hash = 0,
    t_import,
    t_rdql,
    t_end
} test_t;

test_t sel_tests[t_end];

int main(int argc, char *argv[])
{
    RDFSQL *db;
    rs_result *res;
    unsigned int i, j;
    char *optstring = RST_OPTSTR "vi";
    int help = 0;
    int verbosity = 0;
    int c, opt_index = 0;
    int err_count;
    static struct option long_options[] = {
	RST_LONGOPTS,
	{"verbose", 0, 0, 'v'},
	{"import-time", 0, 0, 'i'},
	{0, 0, 0, 0}
    };

    rst_set_db("_test");

    while ((c = getopt_long (argc, argv, optstring, long_options, &opt_index)) != -1) {
        if (!rst_handle_option(c, optarg)) {
            if (c == 'v') {
                verbosity++;
            } else {
                help++;
            }
        }
    }

    if (help) {
        printf("Usage: %s "RST_OPTIONS" [-v] [test commands]\n", argv[0]);
        printf(RST_HELP);
        printf("   -v, --verbose     increase verbosity\n\n");
	printf("test commands: (default: all)\n");
        printf("   h   hash test\n");
        printf("   i   import test\n");
        printf("   r   rdql test\n");
        return 1;
    }

    if (optind == argc) {
	for (i = 0; i < t_end; i++) {
	    sel_tests[i] = 1;
	}
    } else {
	for (i = 0; i < t_end; i++) {
	    sel_tests[i] = 0;
	}
	for (i = optind; i < argc; i++) {
	    char *pos;

	    for (pos = argv[i]; *pos; pos++) {
		switch (*pos) {
		case 'h':
		    sel_tests[t_hash] = 1;
		    break;
		case 'i':
		    sel_tests[t_import] = 1;
		    break;
		case 'r':
		    sel_tests[t_import] = 1;
		    sel_tests[t_rdql] = 1;
		    break;
		default:
		    fprintf(stderr, "unknown test: %c\n", *pos);
		    exit(1);
		}
	    }
	}
    }

    db = rst_try_connect(NULL);

    if (!db) {
	fprintf(stderr, "Could not connect to database, is mysql running?\n");
	exit(1);
    }

    /* try to remove old database, doesnt really matter if it fails */
    system(BIN_DIR "/tstore_setup -e -d _test");

    /* try to create database for test, matters if it fails */
    if (system(BIN_DIR "/tstore_setup -d _test")) {
	fprintf(stderr, "cannot create test database, exiting\n");
	fprintf(stderr, "check tstore_setup has been installed in "
			BIN_DIR "\n");
	exit(1);
    }
   
    #define HTEST(s, val) \
	if (rs_hash(s) != val) { \
	    fprintf(stderr, "failed: hash(\"" s "\")\n"); \
	    exit(1); \
	}
  

    if (sel_tests[t_hash]) {
	HTEST("abc", 0x900150983cd24fb0ULL);  
	HTEST("http://www.w3.org/1999/02/22-rdf-syntax-ns#",
	      0xc846f0e16d37e29aULL);  
	HTEST("http://www.w3.org/2000/01/rdf-schema#", 0xaf09014d2c6b81afULL);  
	HTEST("http://www.w3.org/2002/07/owl#", 0xeb179d48fe507185ULL);  

	printf("passed hash test\n");
    }

    if (sel_tests[t_import]) {
	struct timeval then, now;
	long int seconds;
	double duration;

	gettimeofday(&then, NULL);
	if (rs_import(db, "file:" PKG_DATA_DIR "/test-a.rdf", "file:test-a.rdf", 0)) {
	    FAIL("cannot load file:" PKG_DATA_DIR "/test-a.rdf");
	}
	if (rs_import(db, "file:" PKG_DATA_DIR "/test-b.rdf", "file:test-b.rdf", 0)) {
	    FAIL("cannot load file:" PKG_DATA_DIR "/test-b.rdf");
	}
	gettimeofday(&now, NULL);
	seconds = now.tv_sec - then.tv_sec;
	duration = (double)seconds + (now.tv_usec - then.tv_usec) * 0.000001;
	
	printf("rs_import() of 2 files, 6800 triples took %fs\n", duration);

	if (rs_import(db, "file:" PKG_DATA_DIR "/test-c.nt", "file:test-c.nt", 0)) {
	    FAIL("cannot load file:" PKG_DATA_DIR "/test-c.nt");
	}

	res = rs_find_asserted_triples(db, NULL, NULL, NULL, ObjAny, 0,
				       "file:test-c.nt");
	if (rs_num_triples(res) != 6) {
	    FAIL("import of file:test-c.nt returned %d triples, not 6",
		 rs_num_triples(res));
	}

	err_count = 0;
	err_count += res_test(res, "http://example.com/testc#foo",
	    "http://example.com/testc#bar", "http://example.com/testc#baz");
	err_count += res_test(res, "http://example.com/testc#foo",
	    "http://example.com/testc#bar", "http://example.com/testc#qux");
	err_count += res_test(res, "http://example.com/testc#foo",
	    "http://example.com/testc#bar", "'foobar");
	err_count += res_test(res, "http://example.com/testc#foo",
	    "http://example.com/testc#bar", "'bazqux");
	err_count += res_test(res, "http://example.com/testc#a",
	    "http://example.com/testc#b", "http://example.com/testc#c");
	err_count += res_test(res, "http://example.com/testc#c",
	     "http://example.com/testc#d", "'e");
	if (err_count != 0) {
	    FAIL("import of file:test-c.nt failed to produce %d triple(s) that "
		 "should be present", err_count);
	}
	rs_free_result(res);

	printf("passed assertion and ground query test\n");
    }


    #define RDQL_TEST(q) \
	rdql_res = rs_rdql_query(db, q, 1000); \
	if (rdql_res->error_str) { \
	    FAIL("RDQL query " q "\n%s\n", rdql_res->error_str); \
	} \
	rs_rdql_free_result(rdql_res);
		    
    if (sel_tests[t_rdql]) {
	rs_rdql_result *rdql_res;

	RDQL_TEST("SELECT ?p, ?o WHERE (<http://example.com/testc#foo>, ?p, ?o)");
	RDQL_TEST("SELECT ?p ?o WHERE (<http://example.com/testc#foo>, ?p, ?o)");
	RDQL_TEST("SELECT ?p ?o WHERE (<http://example.com/testc#foo> ?p ?o)");
	RDQL_TEST("SELECT ?p ?o WHERE (<c:foo> ?p ?o) USING c FOR <http://example.com/testc#>");
	RDQL_TEST("SELECT ?p ?o WHERE (c:foo ?p ?o) USING c FOR <http://example.com/testc#>");

	RDQL_TEST("SELECT ?p, ?o WHERE "
		  "(<http://example.com/testc#foo>, ?p, ?o)," 
		  "(<http://example.com/testc#foo>, ?p, ?o)");
	RDQL_TEST("SELECT ?p, ?o WHERE "
		  "(<http://example.com/testc#foo>, ?p, ?o)" 
		  "(<http://example.com/testc#foo>, ?p, ?o)");
	RDQL_TEST("SELECT ?p ?o WHERE "
		  "(<http://example.com/testc#foo> ?p ?o)" 
		  "(<http://example.com/testc#foo>, ?p, ?o)");

	printf("passed RDQL syntax query tests\n");
    }

#if 0
    rs_assert_triple(db, "foo", "bar", "baz", ObjURI);
    rs_assert_triple(db, "foo", "quz", "This is a string", ObjLiteral);

    /*res = rs_find_triples(db, NULL, "http://www.w3.org/1999/02/22-rdf-syntax-ns#type", NULL, ObjURI); */
    /*res = rs_find_triples(db, NULL, "http://www.aktors.org/ontology/portal#family-name", "Smith", ObjLiteral, 0, NULL); */
    res = rs_find_triples(db, NULL, "foo", "Smith", ObjLiteral, 0, NULL);
    while ((t = rs_next_triple(res))) {
	printf("%s -- %s -> %s\n", t->subject, t->predicate, t->object);
    }
    rs_free_result(res);

    res = rs_find_triples(db, "Id:239790", NULL, NULL, ObjAny, 0, NULL);
    while ((t = rs_next_triple(res))) {
	printf("%s -- %s -> %s\n", t->subject, t->predicate, t->object);
    }
    rs_free_result(res);

    for (j = 0; j < 20; j++) {
	rdtscll(then);
	rdql_res =
	    rs_rdql_query(db, "SELECT ?p, ?o WHERE (<foo>, ?p, ?o)", 1000);
	rdtscll(now);
	if (rdql_res->error_str) {
	    printf("error: %s\n", rdql_res->error_str);
	    return 1;
	}
	printf("RDQL query took %lld cycles\n", now - then);
	while ((row = rs_rdql_next_row(rdql_res))) {
	    for (i = 0; i < rdql_res->columns; i++) {
		if (row[i].value)
		    printf("%s\t", row[i].value);
	    }
	    printf("\n");
	}
	rs_rdql_free_result(rdql_res);
    }
    for (j = 0; j < 20; j++) {
	rdtscll(then);
	rdql_res =
	    rs_rdql_query(db,
			  "SELECT ?p, ?o WHERE (?s, ?p, ?o), (?s, <rdf:type>, <akt:Person>), (?s, <rdfs:label>, \"Nigel Shadbolt\") USING akt FOR <http://www.aktors.org/ontology/portal#>, rdfs FOR <http://www.w3.org/2000/01/rdf-schema#>, rdf FOR <http://www.w3.org/1999/02/22-rdf-syntax-ns#>",
			  1000);
	rdtscll(now);
	if (rdql_res->error_str) {
	    printf("error: %s\n", rdql_res->error_str);
	    return 1;
	}
	printf("RDQL query took %lld cycles\n", now - then);
	while ((row = rs_rdql_next_row(rdql_res))) {
	    /* do nothing */
	}
	rs_rdql_free_result(rdql_res);
    }
#endif
    rs_flush(db);

    return 0;
}

int res_test(rs_result *res, const char *s, const char *p, const char *o)
{
    rs_triple *t;

    mysql_data_seek((MYSQL_RES *)res->db_res, 0);

    while ((t = rs_next_triple(res))) {
	if (*o != '\'' && !strcmp(s, t->subject) && !strcmp(p, t->predicate) &&
	    !strcmp(o, t->object)) {
	    return 0;
	}
	if (*o == '\'' && !strcmp(s, t->subject) && !strcmp(p, t->predicate) &&
	    !strcmp(o+1, t->object)) {
	    return 0;
	}
    }

    return 1;
}

/* vi:set ts=8 sts=4 sw=4: */
